/**
 * Copyright (C) 2012-2016  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dtable.client;

import static com.google.gwt.query.client.GQuery.$;
import static com.google.gwt.query.client.GQuery.window;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.dussan.vaadin.dtable.client.components.Filter;
import org.dussan.vaadin.dtable.client.components.LoadingMessage;
import org.dussan.vaadin.dtable.client.components.Order;
import org.dussan.vaadin.dtable.client.components.Pager;
import org.dussan.vaadin.dtable.client.components.QuickSearch;
import org.dussan.vaadin.dtable.client.components.Search;
import org.dussan.vaadin.dtable.client.components.Splitter;
import org.dussan.vaadin.dtable.client.components.TableTools;
import org.dussan.vaadin.dtable.client.components.events.TableEvent;
import org.dussan.vaadin.dtable.client.components.menu.ContextMenu;
import org.dussan.vaadin.dtable.client.components.menu.TableToolsMenu;
import org.dussan.vaadin.dtable.client.components.table.TableBody;
import org.dussan.vaadin.dtable.client.components.table.TableHeader;
import org.dussan.vaadin.dtable.client.events.TableEvents;

import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.Style.Cursor;
import com.google.gwt.dom.client.Style.Display;
import com.google.gwt.dom.client.Style.Unit;
import com.google.gwt.event.logical.shared.HasValueChangeHandlers;
import com.google.gwt.event.logical.shared.ValueChangeEvent;
import com.google.gwt.event.logical.shared.ValueChangeHandler;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.query.client.Function;
import com.google.gwt.query.client.GQuery;
import com.google.gwt.query.client.css.CSS;
import com.google.gwt.query.client.css.Length;
import com.google.gwt.query.client.css.TakesCssValue.CssSetter;
import com.google.gwt.query.client.css.TextAlignProperty.TextAlign;
import com.google.gwt.user.client.DOM;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.Timer;
import com.google.gwt.user.client.Window;
import com.vaadin.client.BrowserInfo;
import com.vaadin.client.ui.VCustomLayout;

public class VDTable
    extends VCustomLayout
    implements HasValueChangeHandlers<Object[]>
{

    public static final TextAlign[] COLUMN_ALIGN = { TextAlign.LEFT, TextAlign.RIGHT, TextAlign.CENTER };

    private static final String ARRAY_DELIMITER = "___:_:___";

    private static final String ARRAY_FIELDS_DELIMITER = "___::___";

    private Boolean fetchNewData = null; // don't change

    private boolean allColumnsVisible = false;

    private boolean allRowsVisible = false;

    private boolean isAnyFrozenColumnHidden = false;

    private boolean frozenTableHidden = false;

    private boolean resizeTimerEnabled = false;

    private boolean scrollTimerEnabled = false;

    private boolean sendingMessage = false;

    private int allColumns = 0;

    private int allRows = 0;

    private int browserOffset = 0;

    private int controledScroll = 0;

    private int ffMarginTop = 0;

    private int firstRow = 0;

    private int left = 0;

    private int scrollLeft = 0;

    private int scrollLeftFrozen = 0;

    private int scrollTop = 0;

    private int scrollTopFirstVisibleRow = 0;

    private int tableFrozenWidth = 0;

    private int tableWidth = 0;

    private int top = 0;

    private int windowX = 0;

    private int windowY = 0;

    private long id = 0;

    private Map<String, String> currentTableData = null;

    private Map<String, String> tableData = null;

    private Map<String, String> tableConfigData = null;

    private Map<String, String> tableExportData = null;

    private Map<String, String> translationData = null;

    private Map<Integer, Integer> columnsWidth = new HashMap<>();

    private List<Integer> columnsFrozen = new ArrayList<>();

    private List<Integer> columnsVisible = new ArrayList<>();

    private ContextMenu contextMenu = null;

    private Filter filter = null;

    private LoadingMessage loadingMessage = null;

    private Order order = null;

    private QuickSearch quickSearch = null;

    private Pager pager = null;

    private Search search = null;

    private Splitter splitter = null;

    private TableBody tableBody = null;

    private TableConfig tableConfig = null;

    private TableHeader tableHeader = null;

    private TableExporter tableExporter = null;

    private TableTools tableTools = null;

    private TableToolsMenu tableToolsMenu = null;

    private Translation translation = null;

    private GQuery frozenTableScrollEventHandler = null;

    private GQuery mouseLeaveWindowEventHandler = null;

    private GQuery tableOutEventHandler = null;

    private GQuery tableScrollEventHandler = null;

    private GQuery windowResizeEventHandler = null;

    public VDTable()
    {
        // global id
        id = new Date().getTime();
        getElement().setId( "v-dtable" + id );
        setStyleName( "v-dtable" );

        currentTableData = new HashMap<>();
        tableConfigData = new HashMap<>();
        tableExportData = new HashMap<>();
        translationData = new HashMap<>();

        translation = new Translation();
        tableConfig = new TableConfig();
        tableExporter = new TableExporter();

        tableHeader = new TableHeader( this );
        tableBody = new TableBody( this );
        splitter = new Splitter( this );
        tableTools = new TableTools( this );
        quickSearch = new QuickSearch( this );
        search = new Search( this );
        order = new Order( this );
        filter = new Filter( this );
        loadingMessage = new LoadingMessage( this );
        pager = new Pager( this );
    }

    public static String[] toStringArray( String value )
    {
        if ( value != null && !value.isEmpty() && value.contains( "[" ) && value.contains( "]" ) )
        {
            String tempValue = value.replace( "[", "" )
                                    .replace( "]", "" );

            String[] values;
            if ( tempValue.contains( ARRAY_DELIMITER ) )
            {
                values = tempValue.split( ARRAY_DELIMITER );
            }
            else
            {
                values = tempValue.split( "," );
            }

            List<String> strings = new ArrayList<>();
            for ( String string : values )
            {
                strings.add( string.trim() );
            }
            return strings.toArray( new String[strings.size()] );
        }
        return new String[] {};
    }

    @Override
    public HandlerRegistration addValueChangeHandler( ValueChangeHandler<Object[]> handler )
    {
        return addHandler( handler, ValueChangeEvent.getType() );
    }

    private void initContainers()
    {
        browserOffset = BrowserInfo.get()
                                   .isChrome() ? 0 : 1;

        if ( $( getElement() ).find( "#dt-main-container-wrapper" )
                              .isEmpty() )
        {
            Element mainContainerWrapper = DOM.createDiv();
            mainContainerWrapper.setId( "dt-main-container-wrapper" );
            getElement().appendChild( mainContainerWrapper );

            Element mainContainer = DOM.createTable();
            mainContainer.setId( "dt-main-container" );
            mainContainerWrapper.appendChild( mainContainer );

            Element message = DOM.createDiv();
            message.setId( "dt-message" + getId() );
            message.setClassName( "dt-message" );
            $( message ).css( CSS.DISPLAY.with( Display.NONE ) );
            mainContainerWrapper.appendChild( message );

            // header section
            Element trHeader = DOM.createTR();
            mainContainer.appendChild( trHeader );

            Element headerWrapper = DOM.createTD();
            headerWrapper.setId( "dt-header-wrapper" );
            trHeader.appendChild( headerWrapper );

            Element headerContainer = DOM.createDiv();
            headerContainer.setId( "dt-header-container" );
            headerWrapper.appendChild( headerContainer );

            Element headerTable = DOM.createTable();
            headerTable.setId( "dt-header-table" + getId() );
            headerTable.getStyle()
                       .setWidth( 100, Unit.PCT );
            headerContainer.appendChild( headerTable );

            Element headerContainerLength = DOM.createDiv();
            headerContainerLength.setId( "dt-header-container-length" + getId() );
            headerContainerLength.setClassName( "dt-header-container-length" );
            headerContainer.appendChild( headerContainerLength );

            Element headerWrapperEmpty = DOM.createTD();
            headerWrapperEmpty.setId( "dt-header-wrapper-empty" );
            trHeader.appendChild( headerWrapperEmpty );

            // body section
            Element trBody = DOM.createTR();
            mainContainer.appendChild( trBody );

            Element bodyWrapper = DOM.createTD();
            bodyWrapper.setId( "dt-body-wrapper" );
            bodyWrapper.setPropertyInt( "colSpan", 2 );
            trBody.appendChild( bodyWrapper );

            Element bodyContainer = DOM.createDiv();
            bodyContainer.setId( "dt-body-container" );
            bodyWrapper.appendChild( bodyContainer );

            Element bodyTable = DOM.createTable();
            bodyTable.setId( "dt-body-table" + getId() );
            bodyTable.getStyle()
                     .setWidth( 100, Unit.PCT );
            bodyContainer.appendChild( bodyTable );

            Element bodyContainerLength = DOM.createDiv();
            bodyContainerLength.setId( "dt-body-container-length" + getId() );
            bodyContainerLength.setClassName( "dt-body-container-length" );
            bodyContainer.appendChild( bodyContainerLength );

            if ( contextMenu == null )
            {
                contextMenu = new ContextMenu( this );
            }

            if ( tableToolsMenu == null )
            {
                tableToolsMenu = new TableToolsMenu( this );
            }
        }
    }

    private void appendFrozenTable()
    {
        // header section
        if ( $( "#dt-frozen-header-wrapper" ).isEmpty() )
        {
            Element frozenHeaderWrapper = DOM.createTD();
            frozenHeaderWrapper.setId( "dt-frozen-header-wrapper" );

            Element frozenHeaderContainer = DOM.createDiv();
            frozenHeaderContainer.setId( "dt-frozen-header-container" );
            frozenHeaderWrapper.appendChild( frozenHeaderContainer );

            Element frozenHeaderTable = DOM.createTable();
            frozenHeaderTable.setId( "dt-frozen-header-table" + getId() );
            frozenHeaderTable.getStyle()
                             .setWidth( 100, Unit.PCT );
            frozenHeaderContainer.appendChild( frozenHeaderTable );

            Element splitterContainer = DOM.createTD();
            splitterContainer.setId( "dt-splitter-container" );
            splitterContainer.setAttribute( "rowspan", "2" );
            splitterContainer.getStyle()
                             .setCursor( Cursor.DEFAULT );

            $( "table#dt-header-table" + getId() ).parent()
                                                  .parent()
                                                  .before( $( frozenHeaderWrapper ) );
            $( "table#dt-header-table" + getId() ).parent()
                                                  .parent()
                                                  .before( $( splitterContainer ) );
        }

        // body section
        if ( $( "#dt-frozen-body-wrapper" ).isEmpty() )
        {
            Element frozenBodyWrapper = DOM.createTD();
            frozenBodyWrapper.setId( "dt-frozen-body-wrapper" );

            Element frozenBodyContainer = DOM.createDiv();
            frozenBodyContainer.setId( "dt-frozen-body-container" );
            frozenBodyWrapper.appendChild( frozenBodyContainer );

            Element frozenBodyTable = DOM.createTable();
            frozenBodyTable.setId( "dt-frozen-body-table" + getId() );
            frozenBodyTable.getStyle()
                           .setWidth( 100, Unit.PCT );
            frozenBodyContainer.appendChild( frozenBodyTable );

            Element frozenBodyContainerLength = DOM.createDiv();
            frozenBodyContainerLength.setId( "dt-frozen-body-container-length" + getId() );
            frozenBodyContainerLength.setClassName( "dt-frozen-body-container-length" );
            frozenBodyContainer.appendChild( frozenBodyContainerLength );

            $( "table#dt-body-table" + getId() ).parent()
                                                .parent()
                                                .before( $( frozenBodyWrapper ) );
        }
    }

    private void loadComponents()
    {
        tableHeader.load();
        tableBody.load();
        splitter.load();
        tableTools.load();
        quickSearch.load();
        search.load();
        order.load();
        filter.load();
        loadingMessage.load();
        pager.load();
    }

    private void unloadComponents()
    {
        tableHeader.unload();
        tableBody.unload();
        splitter.unload();
        tableTools.unload();
        quickSearch.unload();
        search.unload();
        order.unload();
        filter.unload();
        loadingMessage.unload();
        pager.unload();
    }

    private void activateWindowResizeEventHandler( boolean activate )
    {
        if ( activate )
        {
            activateWindowResizeEventHandler( false );
            windowResizeEventHandler = $( window ).resize( new Function()
            {
                public void f()
                {
                    $( "div#dt-header-container-length" + getId() ).css( CSS.WIDTH.with( Length.px( tableWidth ) ) );
                    $( "div#dt-body-container-length" + getId() ).css( CSS.WIDTH.with( Length.px( tableWidth ) ) );

                    if ( !resizeTimerEnabled )
                    {
                        updateTableOnResizing();
                    }

                    if ( config().isTooltipEnabled() )
                    {
                        header().tooltip()
                                .hide();
                    }
                };
            } );
        }
        else if ( windowResizeEventHandler != null )
        {
            windowResizeEventHandler.off();
            windowResizeEventHandler = null;
        }
    }

    private void activateMouseLeaveWindowEventHandler( boolean activate )
    {
        if ( activate )
        {
            activateMouseLeaveWindowEventHandler( false );
            mouseLeaveWindowEventHandler = $( window ).mouseout( new Function()
            {
                public boolean f( Event e )
                {
                    TableEvent.mouseLeaveWindow( $( "table#dt-frozen-header-table" + getId() ),
                                                 $( "table#dt-frozen-header-table" + getId() ),
                                                 $( "table#dt-frozen-body-table" + getId() ),
                                                 $( "table#dt-body-table" + getId() ),
                                                 config().isHorizontalLinesEnabled() );

                    if ( config().isTooltipEnabled() )
                    {
                        header().tooltip()
                                .hide();
                    }

                    return false;
                }
            } );
        }
        else if ( mouseLeaveWindowEventHandler != null )
        {
            mouseLeaveWindowEventHandler.off();
            mouseLeaveWindowEventHandler = null;
        }
    }

    private void activateTableOutEventHandler( boolean activate )
    {
        if ( activate )
        {
            activateTableOutEventHandler( false );
            tableOutEventHandler = $( "div#v-dtable" + getId() ).mouseout( new Function()
            {
                public void f()
                {
                    TableEvent.mouseLeaveWindow( $( "table#dt-frozen-header-table" + getId() ),
                                                 $( "table#dt-frozen-header-table" + getId() ),
                                                 $( "table#dt-frozen-body-table" + getId() ),
                                                 $( "table#dt-body-table" + getId() ),
                                                 config().isHorizontalLinesEnabled() );

                    if ( config().isTooltipEnabled() )
                    {
                        header().tooltip()
                                .hide();
                    }
                };
            } );
        }
        else if ( tableOutEventHandler != null )
        {
            tableOutEventHandler.off();
            tableOutEventHandler = null;
        }
    }

    private void activateTableScrollEventHandler( boolean activate )
    {
        if ( activate )
        {
            final GQuery header = $( "table#dt-header-table" + getId() ).parent();
            final GQuery body = $( "table#dt-body-table" + getId() ).parent();
            final GQuery frozenBody = $( "table#dt-frozen-body-table" + getId() );

            activateTableScrollEventHandler( false );
            tableScrollEventHandler = body.scroll( new Function()
            {
                public void f()
                {
                    header.scrollLeft( body.scrollLeft() );
                    if ( BrowserInfo.get()
                                    .isFirefox() )
                    {
                        // FF bug: flickering if vertical scroll bar is hidden
                        frozenBody.css( CSS.MARGIN_TOP.with( Length.px( config().isPageableTableEnabled() ? 0
                                        : ( -body.scrollTop() + ffMarginTop ) ) ) );
                    }
                    else
                    {
                        frozenBody.parent()
                                  .scrollTop( body.scrollTop() );
                    }

                    if ( !scrollTimerEnabled )
                    {
                        updateTableOnScrolling();
                    }
                };
            } );
        }
        else if ( tableScrollEventHandler != null )
        {
            tableScrollEventHandler.off();
            tableScrollEventHandler = null;
        }
    }

    private void activateFrozenTableScrollEventHandler( boolean activate )
    {
        if ( activate )
        {
            final GQuery header = $( "table#dt-frozen-header-table" + getId() ).parent();
            final GQuery body = $( "table#dt-frozen-body-table" + getId() ).parent();
            activateFrozenTableScrollEventHandler( false );
            frozenTableScrollEventHandler = body.scroll( new Function()
            {
                public void f()
                {
                    scrollLeftFrozen = body.scrollLeft();
                    header.scrollLeft( scrollLeftFrozen );
                };
            } );
        }
        else if ( frozenTableScrollEventHandler != null )
        {
            frozenTableScrollEventHandler.off();
            frozenTableScrollEventHandler = null;
        }
    }

    private void updateTableOnResizing()
    {
        if ( windowX != Window.getClientWidth() || windowY != Window.getClientHeight() )
        {
            resizeTimerEnabled = true;

            Timer timer = new Timer()
            {
                public void run()
                {
                    int winX = Window.getClientWidth();
                    int winY = Window.getClientHeight();
                    GQuery frozenBody = $( "table#dt-frozen-body-table" + getId() );
                    GQuery body = $( "table#dt-body-table" + getId() );

                    if ( windowX == winX && windowY == winY )
                    {
                        // check if table width is smaller then table default minimal width
                        if ( config().isFrozenTableEnabled() && isFrozenTableVisible() //
                            && body.parent()
                                   .outerWidth() < config().getTableMinimalWidth() )
                        {
                            String[][] filters = { { "td#dt-frozen-header-wrapper", "td#dt-frozen-body-wrapper" },
                                { "div#dt-frozen-header-container", "div#dt-frozen-body-container" } };
                            GQuery mainContainerWrapper =
                                $( "div#v-dtable" + getId() ).find( "div#dt-main-container-wrapper" );

                            mainContainerWrapper.children( "table" )
                                                .children( "tr" )
                                                .children( filters[0] )
                                                .css( CSS.WIDTH.with( Length.px( config().getTableFrozenMinimalWidth() ) ) );

                            mainContainerWrapper.children( "table" )
                                                .children( "tr" )
                                                .children( "td" )
                                                .children( filters[1] )
                                                .css( CSS.WIDTH.with( Length.px( config().getTableFrozenMinimalWidth() ) ) );
                        }

                        cancel();
                        resizeTimerEnabled = false;
                        checkContainerDimensionsAndTableCellAlignments( frozenBody, body );
                    }

                    windowX = Window.getClientWidth();
                    windowY = Window.getClientHeight();
                }
            };
            timer.scheduleRepeating( config().getPageResizingActionDelay() );
        }

    }

    private void updateTableOnScrolling()
    {
        final GQuery body = $( "table#dt-body-table" + getId() ).parent();
        if ( scrollLeft != body.scrollLeft() || scrollTop != body.scrollTop() )
        {
            scrollTimerEnabled = true;

            Timer timer = new Timer()
            {
                public void run()
                {
                    int leftScroll = body.scrollLeft();
                    int topScroll = body.scrollTop();

                    if ( scrollLeft == leftScroll && scrollTop == topScroll )
                    {
                        if ( !config().isAllRowsEnabled() || !config().isAllColumnsEnabled() )
                        {
                            fetchNewData = scrollLeft == controledScroll;
                            controledScroll = scrollLeft;
                            if ( ( fetchNewData && !allRowsVisible ) || ( !fetchNewData && !allColumnsVisible ) )
                            {
                                sendMessage( "newData", fetchNewData.toString(), true );
                            }
                        }

                        cancel();
                        scrollTimerEnabled = false;
                        fetchNewData = null;
                    }

                    scrollLeft = leftScroll;
                    scrollTop = topScroll;
                }
            };
            timer.scheduleRepeating( config().getScrollbarActionDelay() );
        }
    }

    public void setEventsHandlerEnabled()
    {
        tableHeader.setEventsHandlerEnabled();
        tableBody.setEventsHandlerEnabled();
        splitter.setEventsHandlerEnabled();
        tableTools.setEventsHandlerEnabled();
        quickSearch.setEventsHandlerEnabled();
        search.setEventsHandlerEnabled();
        order.setEventsHandlerEnabled();
        filter.setEventsHandlerEnabled();
        loadingMessage.setEventsHandlerEnabled();
        pager.setEventsHandlerEnabled();
        sendingMessage = false;
    }

    public void setEventsHandlerDisabled()
    {
        sendingMessage = true;
        tableHeader.setEventsHandlerDisabled();
        tableBody.setEventsHandlerDisabled();
        splitter.setEventsHandlerDisabled();
        tableTools.setEventsHandlerDisabled();
        quickSearch.setEventsHandlerDisabled();
        search.setEventsHandlerDisabled();
        order.setEventsHandlerDisabled();
        filter.setEventsHandlerDisabled();
        loadingMessage.setEventsHandlerDisabled();
        pager.setEventsHandlerDisabled();
    }

    public Translation translation()
    {
        return translation;
    }

    public TableConfig config()
    {
        return tableConfig;
    }

    public TableHeader header()
    {
        return tableHeader;
    }

    public TableExporter exporter()
    {
        return tableExporter;
    }

    public TableTools tools()
    {
        return tableTools;
    }

    public TableToolsMenu tableToolsMenu()
    {
        return tableToolsMenu;
    }

    public QuickSearch quickSearch()
    {
        return quickSearch;
    }

    public Search search()
    {
        return search;
    }

    public Order order()
    {
        return order;
    }

    public Filter filter()
    {
        return filter;
    }

    public ContextMenu contextMenu()
    {
        return contextMenu;
    }

    public long getId()
    {
        return id;
    }

    public int getAllColumns()
    {
        return allColumns;
    }

    public int getFirstRow()
    {
        return firstRow;
    }

    public int getAllRows()
    {
        return allRows;
    }

    public int getFrozenTableWidth()
    {
        return tableFrozenWidth;
    }

    public void setFrozenTableWidth( int tableFrozenWidth )
    {
        this.tableFrozenWidth = tableFrozenWidth;
    }

    public int getLeft()
    {
        return left;
    }

    public int getScrollLeftFrozen()
    {
        return scrollLeftFrozen;
    }

    public int getScrollLeft()
    {
        return scrollLeft;
    }

    public int getBrowserOffset()
    {
        return browserOffset;
    }

    public List<Integer> getColumnsFrozen()
    {
        return columnsFrozen;
    }

    public List<Integer> getColumnsVisible()
    {
        return columnsVisible;
    }

    public Map<Integer, Integer> getColumnsWidth()
    {
        return columnsWidth;
    }

    public boolean isFrozenTableEmpty()
    {
        return columnsFrozen.isEmpty();
    }

    public boolean isFrozenTableHidden()
    {
        return frozenTableHidden;
    }

    public void hideFrozenTable( boolean hide )
    {
        frozenTableHidden = hide;
        showFrozenTable( !hide );
    }

    public boolean isFrozenTableVisible()
    {
        return !frozenTableHidden;
    }

    public boolean isAllColumnsVisible()
    {
        return allColumnsVisible;
    }

    public boolean isAnyFrozenColumnHidden()
    {
        return isAnyFrozenColumnHidden;
    }

    public void setAllColumnsVisible( boolean allColumnsVisible )
    {
        this.allColumnsVisible = allColumnsVisible;
    }

    public void showFrozenTable( boolean show )
    {
        final String[][] filters = { { "td#dt-frozen-header-wrapper", "td#dt-frozen-body-wrapper" },
            { "div#dt-frozen-header-container", "div#dt-frozen-body-container" } };
        GQuery mainContainer = $( "div#v-dtable" + getId() + " table#dt-main-container" );

        if ( show )
        {
            mainContainer.children( "tr" )
                         .children( filters[0] )
                         .css( CSS.WIDTH.with( Length.px( tableFrozenWidth ) ) );

            mainContainer.children( "tr" )
                         .children( "td" )
                         .children( filters[1] )
                         .css( CSS.WIDTH.with( Length.px( tableFrozenWidth ) ) );

            // set width to 100% - Chrome want this
            mainContainer.children( "tr" )
                         .children( "td" )
                         .children( filters[1] )
                         .css( CSS.WIDTH.with( Length.pct( 100 ) ) );
        }
        else
        {
            mainContainer.children( "tr" )
                         .children( filters[0] )
                         .css( CSS.WIDTH.with( Length.px( 0 ) ) );

            mainContainer.children( "tr" )
                         .children( "td" )
                         .children( filters[1] )
                         .css( CSS.WIDTH.with( Length.px( 0 ) ) );
        }
    }

    public boolean isFrozen( int column )
    {
        return columnsFrozen.contains( column );
    }

    public void setFrozen( int column )
    {
        // remove filter
        if ( ( !config().isFilteredFrozenTableEnabled() && !isFrozen( column ) && filter.isFiltered( column ) )
            || ( !config().isFilteredTableEnabled() && isFrozen( column ) && filter.isFiltered( column ) ) )
        {
            currentTableData.put( "removeFilter", Integer.toString( column ) );
        }

        // remove order
        if ( !config().isOrderedFrozenTableEnabled() && !isFrozen( column ) )
        {
            $( "table#dt-header-table" + getId() ).children( "tr" )
                                                  .children( "td [id='" + getId() + ":" + column + "']" )
                                                  .removeClass( "asc", "desc" );
            currentTableData.put( "order", order.getOrders() );
        }
        else if ( !config().isOrderedTableEnabled() && isFrozen( column ) )
        {
            $( "table#dt-frozen-header-table" + getId() ).children( "tr" )
                                                         .children( "td [id='" + getId() + ":" + column + "']" )
                                                         .removeClass( "asc", "desc" );
            currentTableData.put( "order", order.getOrders() );
        }

        sendMessage( "freeze", Integer.toString( column ), true );
    }

    private void setContainerWidthAndHeight( GQuery table, GQuery virtualTable, CssSetter tableHeight,
                                             CssSetter virtualTableWidth, CssSetter virtualTableHeight )
    {
        table.css( tableHeight );
        virtualTable.css( virtualTableWidth )
                    .css( virtualTableHeight );
    }

    private void setScrollableContainerWidthAndHeight( GQuery virtualHeaderTable, GQuery table, GQuery virtualTable,
                                                       GQuery frozenTable, GQuery frozenVirtualTable,
                                                       Length tableHeight, Length virtualTableWidth,
                                                       Length virtualTableHeight, Length frozenVirtualTableHeight )
    {
        // set header container width and height
        virtualHeaderTable.css( CSS.WIDTH.with( virtualTableWidth ) )
                          .css( CSS.HEIGHT.with( Length.px( virtualHeaderTable.parent()
                                                                              .height() ) ) );

        // set frozen body container width and height
        if ( config().isFrozenTableEnabled() )
        {
            setContainerWidthAndHeight( frozenTable, frozenVirtualTable, CSS.HEIGHT.with( tableHeight ),
                                        CSS.WIDTH.with( Length.pct( 100 ) ),
                                        CSS.HEIGHT.with( frozenVirtualTableHeight ) );
        }

        // set body container width and height
        setContainerWidthAndHeight( table, virtualTable, CSS.HEIGHT.with( tableHeight ),
                                    CSS.WIDTH.with( virtualTableWidth ), CSS.HEIGHT.with( virtualTableHeight ) );
    }

    private void setPageableContainerWidthAndHeight( GQuery virtualHeaderTable, GQuery table, GQuery virtualTable,
                                                     GQuery frozenTable, GQuery frozenVirtualTable, Length tableHeight,
                                                     Length virtualTableWidth )
    {
        // set header container width and height
        virtualHeaderTable.css( CSS.WIDTH.with( virtualTableWidth ) )
                          .css( CSS.HEIGHT.with( Length.px( virtualHeaderTable.parent()
                                                                              .height() ) ) );

        // set frozen body container width and height
        if ( config().isFrozenTableEnabled() )
        {
            setContainerWidthAndHeight( frozenTable, frozenVirtualTable, CSS.HEIGHT.with( tableHeight ),
                                        CSS.WIDTH.with( Length.pct( 100 ) ), CSS.HEIGHT.with( Length.pct( 100 ) ) );
        }

        // set body container width and height
        setContainerWidthAndHeight( table, virtualTable, CSS.HEIGHT.with( tableHeight ),
                                    CSS.WIDTH.with( virtualTableWidth ), CSS.HEIGHT.with( Length.pct( 100 ) ) );
    }

    private void setNonfixedContainerWidthAndHeight( GQuery virtualHeaderTable, GQuery table, GQuery virtualTable,
                                                     GQuery frozenTable, GQuery frozenVirtualTable,
                                                     Length virtualTableWidth )
    {
        // set header container width and height
        virtualHeaderTable.css( CSS.WIDTH.with( virtualTableWidth ) )
                          .css( CSS.HEIGHT.with( Length.px( virtualHeaderTable.parent()
                                                                              .height() ) ) );

        // set frozen body container width and height
        if ( config().isFrozenTableEnabled() )
        {
            setContainerWidthAndHeight( frozenTable, frozenVirtualTable, CSS.HEIGHT.with( Length.pct( 100 ) ),
                                        CSS.WIDTH.with( Length.pct( 100 ) ), CSS.HEIGHT.with( Length.pct( 100 ) ) );
        }

        // set body container width and height
        setContainerWidthAndHeight( table, virtualTable, CSS.HEIGHT.with( Length.pct( 100 ) ),
                                    CSS.WIDTH.with( virtualTableWidth ), CSS.HEIGHT.with( Length.pct( 100 ) ) );
    }

    private void setContainerWidthAndHeight()
    {
        // calculate horizontal scrollbar height
        GQuery body = $( "table#dt-body-table" + getId() ).parent();
        int tableHeight = allRows * config().getBodyRowHeight();
        int frozenScrollbarHeight = 0;
        if ( config().isFrozenTableEnabled() && isFrozenTableVisible() )
        {
            GQuery frozenBody = $( "table#dt-frozen-body-table" + getId() ).parent();
            frozenScrollbarHeight = frozenBody.outerHeight() - frozenBody.height();
        }

        int scrollbarHeight = body.outerHeight() - body.height();
        if ( !config().isPageableTableEnabled()
            && ( config().isFixedTableHeightEnabled() || config().getTableHeight() <= tableHeight ) )
        {
            // calculate scrollbar's addition to body container height
            int frozenTableHeightOffset = Math.max( 0, scrollbarHeight - frozenScrollbarHeight );
            int tableHeightOffset = Math.max( 0, frozenScrollbarHeight - scrollbarHeight );
            frozenTableHeightOffset += frozenTableHeightOffset > 0 ? browserOffset : 0;
            tableHeightOffset += tableHeightOffset > 0 ? browserOffset : 0;

            setScrollableContainerWidthAndHeight( $( "div#dt-header-container-length" + getId() ),
                                                  $( "table#dt-body-table" + getId() ).parent(),
                                                  $( "div#dt-body-container-length" + getId() ),
                                                  $( "table#dt-frozen-body-table" + getId() ).parent(),
                                                  $( "div#dt-frozen-body-container-length" + getId() ),
                                                  Length.px( config().getTableHeight() ), // tableHeight
                                                  allColumnsVisible ? Length.pct( 100 ) : Length.px( tableWidth ), // virtualTableWidth
                                                  Length.px( config().getBodyRowHeight() * allRows
                                                      + tableHeightOffset ), // virtualTableHeight
                                                  Length.px( config().getBodyRowHeight() * allRows
                                                      + frozenTableHeightOffset ) ); // frozenVirtualTableHeight
        }
        else if ( config().isPageableTableEnabled() && config().isFixedTableHeightEnabled() && pager.getAllPages() > 1 )
        {
            setPageableContainerWidthAndHeight( $( "div#dt-header-container-length" + getId() ),
                                                $( "table#dt-body-table" + getId() ).parent(),
                                                $( "div#dt-body-container-length" + getId() ),
                                                $( "table#dt-frozen-body-table" + getId() ).parent(),
                                                $( "div#dt-frozen-body-container-length" + getId() ),
                                                Length.px( config().getTableHeight()
                                                    + Math.max( frozenScrollbarHeight, scrollbarHeight ) + 1 ), // virtualTableHeight
                                                allColumnsVisible ? Length.pct( 100 ) : Length.px( tableWidth ) ); // virtualTableWidth
        }
        else
        {
            setNonfixedContainerWidthAndHeight( $( "div#dt-header-container-length" + getId() ),
                                                $( "table#dt-body-table" + getId() ).parent(),
                                                $( "div#dt-body-container-length" + getId() ),
                                                $( "table#dt-frozen-body-table" + getId() ).parent(),
                                                $( "div#dt-frozen-body-container-length" + getId() ),
                                                allColumnsVisible ? Length.pct( 100 ) : Length.px( tableWidth ) ); // virtualTableWidth
        }

        // add vertical scrollbar's width to empty header wrapper
        int width = body.outerWidth() - body.width();
        width -= width > 1 ? 1 : 0;
        $( "div#v-dtable" + getId() + " td#dt-header-wrapper-empty" ).css( CSS.WIDTH.with( Length.px( width ) ) );
    }

    public void checkTableScrollers( GQuery frozenHeader, GQuery frozenBody, GQuery header, GQuery body )
    {
        // set body container width and height
        setContainerWidthAndHeight();

        scrollTopFirstVisibleRow = -scrollTop + top;
        if ( config().isFrozenTableEnabled() )
        {
            frozenHeader.parent()
                        .scrollLeft( scrollLeftFrozen );
            frozenBody.parent()
                      .scrollLeft( scrollLeftFrozen );
            if ( BrowserInfo.get()
                            .isFirefox() )
            {
                // FF bug: flickering if vertical scrollbar is hidden
                frozenBody.css( CSS.MARGIN_TOP.with( Length.px( config().isPageableTableEnabled() ? 0
                                : scrollTopFirstVisibleRow ) ) );
                ffMarginTop = top;
            }
            else
            {
                // set body container position
                frozenBody.css( CSS.TOP.with( Length.px( top ) ) );
                // scroll table body into origin position
                frozenBody.parent()
                          .scrollTop( scrollTop );
            }
        }

        // set header container position
        header.css( CSS.LEFT.with( Length.px( !config().isAllHeaderColumnsEnabled() ? left : 0 ) ) );
        header.parent()
              .scrollLeft( scrollLeft );

        // set body container position
        body.css( CSS.LEFT.with( Length.px( left ) ) );
        body.css( CSS.TOP.with( Length.px( top ) ) );

        // scroll table body into origin position
        body.parent()
            .scrollLeft( scrollLeft );
        body.parent()
            .scrollTop( scrollTop );
    }

    public void checkContainerDimensionsAndTableCellAlignments( GQuery frozenBody, GQuery body )
    {
        setContainerWidthAndHeight();
        tableBody.checkTdHeightsAndAlignments( frozenBody.children( "tr" ), true );
        tableBody.checkTdHeightsAndAlignments( body.children( "tr" ), true );
    }

    private int prepareTable( Map<String, String> data )
    {
        List<Integer> frozenAndVisibleColumns = new ArrayList<>();
        columnsFrozen.clear();
        if ( data.containsKey( "frozenColumns" ) )
        {
            String[] frozenColumns = VDTable.toStringArray( data.get( "frozenColumns" ) );
            for ( String column : frozenColumns )
            {
                columnsFrozen.add( Integer.parseInt( column ) );
            }
            frozenAndVisibleColumns.addAll( columnsFrozen );
        }

        columnsVisible.clear();
        if ( data.containsKey( "visibleColumns" ) )
        {
            String[] visibleColumns = VDTable.toStringArray( data.get( "visibleColumns" ) );
            for ( String column : visibleColumns )
            {
                columnsVisible.add( Integer.parseInt( column ) );
            }
            frozenAndVisibleColumns.addAll( columnsVisible );
        }
        Collections.sort( frozenAndVisibleColumns );

        if ( data.containsKey( "columnsWidth" ) )
        {
            columnsWidth.clear();
            String[] widths = VDTable.toStringArray( data.get( "columnsWidth" ) );
            for ( int column = 0; column < frozenAndVisibleColumns.size(); column++ )
            {
                columnsWidth.put( frozenAndVisibleColumns.get( column ), Integer.parseInt( widths[column] ) );
            }
        }

        if ( config().isFrozenTableEnabled() )
        {
            if ( isFrozenTableEmpty() )
            {
                showFrozenTable( false );
            }
            else if ( !isFrozenTableEmpty() && isFrozenTableVisible() && config().isFrozenTableVisibleEnabled() )
            {
                showFrozenTable( true );
            }
        }

        int newColumnsSize = 0;
        for ( int column : frozenAndVisibleColumns )
        {
            if ( data.containsKey( Integer.toString( column ) ) )
            {
                newColumnsSize += 1;
            }
        }
        return newColumnsSize;
    }

    private void processNewRequest( Map<String, String> currentTableData )
    {
        String newRequest = new String();
        if ( currentTableData != null && !currentTableData.isEmpty() )
        {
            for ( String key : currentTableData.keySet() )
            {
                if ( !newRequest.isEmpty() )
                {
                    newRequest += ARRAY_DELIMITER;
                }
                newRequest += key + ARRAY_FIELDS_DELIMITER + currentTableData.get( key )
                                                                             .trim();
            }
        }

        TableEvents.fireNewRequest( this, newRequest );
    }

    public void processTableData()
    {
        currentTableData.clear();

        // START: this must be checked first
        if ( !translationData.isEmpty() )
        {
            translation().processData( translationData );
        }

        if ( !tableExportData.isEmpty() )
        {
            exporter().processData( tableExportData );
        }

        if ( !tableConfigData.isEmpty() )
        {
            config().processData( tableConfigData );

            if ( tableData.containsKey( "formatedColumns" ) )
            {
                config().setColumnsFormatted( VDTable.toStringArray( tableData.get( "formatedColumns" ) ) );
            }

            if ( tableData.containsKey( "freezeColumnsEnabled" ) )
            {
                config().setColumnsFreezeEnabled( VDTable.toStringArray( tableData.get( "freezeColumnsEnabled" ) ) );
            }

            if ( tableData.containsKey( "unfreezeColumnsEnabled" ) )
            {
                config().setColumnsUnfreezeEnabled( VDTable.toStringArray( tableData.get( "unfreezeColumnsEnabled" ) ) );
            }

            if ( tableData.containsKey( "hideColumnsEnabled" ) )
            {
                config().setColumnsHideEnabled( VDTable.toStringArray( tableData.get( "hideColumnsEnabled" ) ) );
            }

            if ( config().isFrozenTableEnabled() )
            {
                tableFrozenWidth = Math.max( config().getTableFrozenMinimalWidth(), tableFrozenWidth );
                appendFrozenTable();
                showFrozenTable( config().isFrozenTableVisibleEnabled() & isFrozenTableVisible() ); // don't change
                activateFrozenTableScrollEventHandler( true );
                activateTableScrollEventHandler( true ); // don't remove
            }
            else
            {
                tableFrozenWidth = 0;
            }

            if ( !sendingMessage() )
            {
                loadComponents();
            }
        }
        // END: this must be checked first

        if ( tableData.containsKey( "refreshVisibleData" ) )
        {
            if ( tableData.containsKey( "refreshVisibleDataCause" )
                && "hideUnhideColumns".equals( tableData.get( "refreshVisibleDataCause" ) ) )
            {

                sendMessage( "hideColumn", "newData", Boolean.parseBoolean( tableData.get( "refreshVisibleData" ) ) );
            }
            else
            {
                sendMessage( "newData", Boolean.FALSE.toString(),
                             Boolean.parseBoolean( tableData.get( "refreshVisibleData" ) ) );
            }
        }
        else if ( tableData.containsKey( "emptyData" ) )
        {
            // no data to display
            getElement().setInnerHTML( translation().get( Translation.NO_DATA_TO_DISPLAY ) );
        }
        else if ( tableData.containsKey( "noData" ) )
        {
            if ( "searchError".equals( tableData.get( "noData" ) ) )
            {
                String[] values = VDTable.toStringArray( tableData.get( "searchError" ) );
                search().markSearchField( Integer.parseInt( values[0] ) );
                loadingMessage.hide();
            }
            else
            {
                if ( tableData.containsKey( "emptyQuickSearch" ) )
                {
                    quickSearch().emptyQuickSearch();
                }
                if ( tableData.containsKey( "emptySearch" ) )
                {
                    search().emptySearch( Integer.parseInt( tableData.get( "emptySearch" ) ) );
                }
                sendMessage( "noData", "noData", false );
                loadingMessage.hide();
            }
        }
        else if ( tableData.containsKey( "dataError" ) )
        {
            sendMessage( "dataError", "dataError", false );
            loadingMessage.hide();
        }
        else
        {
            if ( tableData.containsKey( "allColumnsVisible" ) )
            {
                allColumnsVisible = Boolean.parseBoolean( tableData.get( "allColumnsVisible" ) );
                tableHeader.set100percentWidth( allColumnsVisible );
                tableBody.set100percentWidth( allColumnsVisible );
            }

            if ( tableData.containsKey( "allRowsVisible" ) )
            {
                allRowsVisible = Boolean.parseBoolean( tableData.get( "allRowsVisible" ) );
            }

            if ( tableData.containsKey( "firstRow" ) )
            {
                firstRow = Integer.parseInt( tableData.get( "firstRow" ) );
            }

            if ( tableData.containsKey( "allColumns" ) )
            {
                allColumns = Integer.parseInt( tableData.get( "allColumns" ) );
                allColumns = allColumns > 0 ? allColumns : 0;
            }

            if ( tableData.containsKey( "allRows" ) && allRows != Integer.parseInt( tableData.get( "allRows" ) ) )
            {
                allRows = Integer.parseInt( tableData.get( "allRows" ) );
                pager.setAllPages( allRows );
            }

            if ( tableData.containsKey( "tableWidth" ) )
            {
                tableWidth = Integer.parseInt( tableData.get( "tableWidth" ) );
            }

            if ( tableData.containsKey( "isAnyFrozenColumnHidden" ) )
            {
                isAnyFrozenColumnHidden = Boolean.parseBoolean( tableData.get( "isAnyFrozenColumnHidden" ) );
            }

            int newTableDataSize = prepareTable( tableData );
            if ( newTableDataSize != ( columnsFrozen.size() + columnsVisible.size() )
                || !tableData.containsKey( "left" ) || !tableData.containsKey( "top" ) )
            {
                // frozen size + non-frozen size != data size
                sendMessage( "transmissionError", "transmissionError", false );
            }
            else if ( tableData.containsKey( "visibleColumns" ) && tableData.containsKey( "left" )
                && tableData.containsKey( "top" ) )
            {
                left = Integer.parseInt( tableData.get( "left" ) );
                top = config().isPageableTableEnabled() ? 0 : Integer.parseInt( tableData.get( "top" ) );

                if ( tableData.containsKey( "scrollLeft" ) && tableData.containsKey( "scrollTop" ) )
                {
                    scrollLeft = Integer.parseInt( tableData.get( "scrollLeft" ) );
                    scrollTop = Integer.parseInt( tableData.get( "scrollTop" ) );
                }

                // table header
                tableHeader.processData( tableData );
                tableHeader.show();

                // add/remove search columns
                search.processData( tableData );
                search.apply();

                // add/remove order
                order.processData( tableData );
                order.apply();

                // add/remove filters
                filter.processData( tableData );
                filter.apply();

                // insert data into table body
                tableBody.processData( tableData );
                tableBody.show();

                // don't change: last check of visibility of frozen table
                if ( config().isFrozenTableEnabled() )
                {
                    showFrozenTable( config().isFrozenTableVisibleEnabled() & isFrozenTableVisible()
                        & !columnsFrozen.isEmpty() );
                }

                // set container dimensions
                setContainerWidthAndHeight();

                // hide loading message
                loadingMessage.hide();

                // page refresh: have to be called last because
                // the table context have to be displayed first
                if ( tableData.containsKey( "pageRefresh" ) )
                {
                    int delay = Integer.parseInt( tableData.get( "pageRefresh" ) );
                    ( new Timer()
                    {
                        public void run()
                        {
                            sendMessage( "newData", Boolean.FALSE.toString(), false );
                        }
                    } ).schedule( delay );
                }
            }
        }
    }

    @Override
    protected void onLoad()
    {
        initContainers();

        activateMouseLeaveWindowEventHandler( true );
        activateWindowResizeEventHandler( true );
        activateTableScrollEventHandler( true );
    }

    @Override
    protected void onUnload()
    {
        activateMouseLeaveWindowEventHandler( false );
        activateWindowResizeEventHandler( false );

        activateTableOutEventHandler( false );
        activateFrozenTableScrollEventHandler( false );
        activateTableScrollEventHandler( false );

        unloadComponents();
    }

    public void sentMessage( final String eventType, final String eventMessage )
    {
        if ( "newFilter".equals( eventType ) || "removeFilter".equals( eventType ) || "order".equals( eventType )
            || "freeze".equals( eventType ) || "quickSearch".equals( eventType ) || "search".equals( eventType )
            || "pageReload".equals( eventType )
            || ( "hideColumn".equals( eventType ) && config().isNewDataWhenHideUnhideColumnEnabled() ) )
        {
            firstRow = 0;
            scrollTop = 0;
            scrollLeft = 0;
            scrollLeftFrozen = 0;
            pager.reset();
        }

        // standard message elements
        int topScroll = !config().isPageableTableEnabled() ? scrollTop
                        : ( ( pager.getCurentPage() - 1 ) * config().getPageSize() * config().getBodyRowHeight() );
        int pageSize = !config().isPageableTableEnabled()
                        ? ( (int) ( config().getTableHeight() / config().getBodyRowHeight() ) + 3 )
                        : config().getPageSize();
        currentTableData.put( "left", Integer.toString( scrollLeft ) );
        currentTableData.put( "top", Integer.toString( topScroll ) );
        currentTableData.put( "width", Integer.toString( BrowserInfo.get()
                                                                    .getScreenWidth() ) );
        currentTableData.put( "pageSize", Integer.toString( pageSize ) );
        currentTableData.put( "bodyRowHeight", Integer.toString( config().getBodyRowHeight() ) );

        // non-standard message elements
        currentTableData.put( eventType, eventMessage );

        // fire new request for sending new data to server
        processNewRequest( currentTableData );

        // clear request
        currentTableData.clear();
    }

    public boolean sendingMessage()
    {
        return sendingMessage;
    }

    public void sendMessage( final String eventType, String eventMessage, boolean showLoadingMessage )
    {
        setEventsHandlerDisabled();
        if ( showLoadingMessage && config().isLoadingMessageEnabled() )
        {
            loadingMessage.show();
            loadingMessage.sendMessage( eventType, eventMessage );
        }
        else
        {
            loadingMessage.hide();
            sentMessage( eventType, eventMessage );
        }
    }

    public Map<String, String> getTableData()
    {
        return tableData;
    }

    public void setTableData( Map<String, String> tableData )
    {
        this.tableData = tableData;
    }

    public Map<String, String> getTableConfigData()
    {
        return tableConfigData;
    }

    public void setTableConfigData( Map<String, String> tableConfigData )
    {
        this.tableConfigData = tableConfigData;
    }

    public Map<String, String> getTableExportData()
    {
        return tableExportData;
    }

    public void setTableExportData( Map<String, String> tableExportData )
    {
        this.tableExportData = tableExportData;
    }

    public Map<String, String> getTranslationData()
    {
        return translationData;
    }

    public void setTranslationData( Map<String, String> translationData )
    {
        this.translationData = translationData;
    }

}
