/**
 * Copyright (C) 2012-2016  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dtable.config;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.dussan.vaadin.dtable.DTable;
import org.dussan.vaadin.dtable.config.formatting.Formatting;
import org.dussan.vaadin.dtable.windows.exporter.base.Charset;
import org.dussan.vaadin.dtable.windows.exporter.base.ColumnDefinition;
import org.dussan.vaadin.dtable.windows.exporter.type.CompressionType;
import org.dussan.vaadin.dtable.windows.exporter.type.ExportingColumnType;
import org.dussan.vaadin.dtable.windows.exporter.version.MsAccessVersion;
import org.dussan.vaadin.dtable.windows.exporter.version.MsExcelVersion;

public class TableExporterConfig
    implements Serializable
{

    private static final long serialVersionUID = -6777208258494068964L;

    private boolean excludeBaseDataEnabled = false;

    private boolean exportCsvFileEnabled = false;

    private boolean exportMdbFileEnabled = false;

    private boolean exportPdfFileEnabled = false;

    private boolean exportXlsFileEnabled = false;

    private boolean formattedDataEnabled = false;

    private boolean rowIndexEnabled = false;

    private boolean prepareDataWithHiddenColumnsEnabled = false;

    private boolean prepareDataWithOrdersEnabled = false;

    private boolean prepareDataWithOriginalValuesForHiddenAndRemovedColumnsEnabled = false;

    private boolean prepareDataWithRemovedColumnsEnabled = false;

    private boolean removeOrdersFromColumnsEnabled = false;

    private boolean selectHiddenColumnsEnabled = false;

    private boolean timestampEnabled = false;

    private int windowRefreshDelay = 2000;

    private String baseName = "dtable";

    private String csvColumnDelimiter = ";";

    private String fileName = "dtable";

    private String tableName = "dtable";

    private String timestampFormat = "yyyyMMdd-HHmmss";

    private transient Object defaultValueForHiddenAndRemovedColumns = 0;

    private transient Map<String, Long> allRows = null;

    private transient Map<String, Map<Integer, ColumnDefinition>> columnsDefinition = null;

    private Charset charset = Charset.UTF_8;

    private CompressionType compressionType = CompressionType.NONE;

    private MsAccessVersion msAccessVersion = MsAccessVersion.V2010;

    private MsExcelVersion msExcelVersion = MsExcelVersion.V2007;

    public TableExporterConfig()
    {
        allRows = new HashMap<>();
        columnsDefinition = new HashMap<>();
    }

    public TableExporterConfig( String baseName )
    {
        this();
        if ( baseName != null && !baseName.isEmpty() )
        {
            this.baseName = baseName;
            this.fileName = baseName;
            this.tableName = baseName;
        }
        createColumnDefinition( this.baseName );
    }

    public static Object[] arrayList2objectArray( List<Object[]> datas )
    {
        try
        {
            List<Object> objects = new ArrayList<>();
            for ( Object[] data : datas )
            {
                objects.addAll( Arrays.asList( data ) );
            }
            return objects.toArray();
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            return new Object[] {};
        }
    }

    public String getBaseName()
    {
        return baseName;
    }

    public long getAllRows( String name )
    {
        if ( allRows.containsKey( name ) )
        {
            return allRows.get( name );
        }
        return 0L;
    }

    public void setAllRows( String name, Long allRows )
    {
        this.allRows.put( name, allRows );
    }

    public String[] getColumnsDefinitionName()
    {
        List<String> names = new ArrayList<>( columnsDefinition.keySet() );
        if ( isExcludeBaseDataEnabled() )
        {
            names.remove( names.indexOf( getBaseName() ) );
        }
        return names.toArray( new String[] {} );
    }

    private void createColumnDefinition( String name )
    {
        columnsDefinition.clear();
        columnsDefinition.put( name, new HashMap<Integer, ColumnDefinition>() );
    }

    public boolean isColumnDefinitionEmpty( String name )
    {
        if ( !columnsDefinition.containsKey( name ) )
        {
            return true;
        }
        return columnsDefinition.get( name )
                                .isEmpty();
    }

    public void removeColumnDefinition( String name )
    {
        columnsDefinition.remove( name );
    }

    public Map<String, Map<Integer, ColumnDefinition>> getColumnsDefinition()
    {
        return columnsDefinition;
    }

    public Map<Integer, ColumnDefinition> getColumnDefinition( String name )
    {
        return columnsDefinition.get( name );
    }

    public String getColumnDefinitionHeader( String name, int column )
    {
        Map<Integer, ColumnDefinition> definition = getColumnDefinition( name );
        if ( definition != null && definition.containsKey( column ) )
        {
            return getColumnDefinition( name ).get( column )
                                              .getHeader();
        }
        return null;
    }

    public Formatting getColumnDefinitionFormatting( String name, int column )
    {
        Map<Integer, ColumnDefinition> definition = getColumnDefinition( name );
        if ( definition != null && definition.containsKey( column ) )
        {
            return getColumnDefinition( name ).get( column )
                                              .getFormatting();
        }
        return null;
    }

    public void setColumnDefinitionFormatting( String name, int column, Formatting formatting )
    {
        if ( columnsDefinition.containsKey( name ) )
        {
            Map<Integer, ColumnDefinition> exported = new HashMap<>();
            if ( !isColumnDefinitionEmpty( name ) )
            {
                exported = columnsDefinition.get( name );
            }

            ColumnDefinition definition = exported.get( column );
            if ( definition != null )
            {
                if ( formatting != null )
                {
                    formatting.setColumn( column );
                }
                exported.put( column,
                              new ColumnDefinition( definition.getHeader(), formatting,
                                                    definition.getExportingColumnType(), definition.getPrecision(),
                                                    definition.getScale() ) );
                columnsDefinition.put( name, exported );
            }
        }
    }

    public ExportingColumnType getColumnDefinitionExportingType( String name, int column )
    {
        Map<Integer, ColumnDefinition> definition = getColumnDefinition( name );
        if ( definition != null && definition.containsKey( column ) )
        {
            return getColumnDefinition( name ).get( column )
                                              .getExportingColumnType();
        }
        return null;
    }

    public void setColumnDefinitionExportingType( String name, int column, ExportingColumnType exportingColumnType )
    {
        if ( columnsDefinition.containsKey( name ) )
        {
            Map<Integer, ColumnDefinition> colDefinition = new HashMap<>();
            if ( !isColumnDefinitionEmpty( name ) )
            {
                colDefinition = columnsDefinition.get( name );
            }

            ColumnDefinition definition = colDefinition.get( column );
            if ( definition != null )
            {
                colDefinition.put( column,
                                   new ColumnDefinition( definition.getHeader(), definition.getFormatting(),
                                                         exportingColumnType, definition.getPrecision(),
                                                         definition.getScale() ) );
                columnsDefinition.put( name, colDefinition );
            }
        }
    }

    public Integer getColumnDefinitionLength( String name, int column )
    {
        Map<Integer, ColumnDefinition> definition = getColumnDefinition( name );
        if ( definition != null && definition.containsKey( column ) )
        {
            return getColumnDefinition( name ).get( column )
                                              .getLength();
        }
        return null;
    }

    public void setColumnDefinitionLength( String name, int column, int length )
    {
        if ( columnsDefinition.containsKey( name ) )
        {
            Map<Integer, ColumnDefinition> colDefinition = new HashMap<>();
            if ( !isColumnDefinitionEmpty( name ) )
            {
                colDefinition = columnsDefinition.get( name );
            }

            ColumnDefinition definition = colDefinition.get( column );
            if ( definition != null )
            {
                colDefinition.put( column,
                                   new ColumnDefinition( definition.getHeader(), definition.getFormatting(),
                                                         definition.getExportingColumnType(), length,
                                                         definition.getScale() ) );
                columnsDefinition.put( name, colDefinition );
            }
        }
    }

    public Integer getColumnDefinitionPrecision( String name, int column )
    {
        Map<Integer, ColumnDefinition> definition = getColumnDefinition( name );
        if ( definition != null && definition.containsKey( column ) )
        {
            return getColumnDefinition( name ).get( column )
                                              .getLength();
        }
        return null;
    }

    public Integer getColumnDefinitionScale( String name, int column )
    {
        Map<Integer, ColumnDefinition> definition = getColumnDefinition( name );
        if ( definition != null && definition.containsKey( column ) )
        {
            return getColumnDefinition( name ).get( column )
                                              .getScale();
        }
        return null;
    }

    public void setColumnDefinitionPrecisionAndScale( String name, int column, int precision, int scale )
    {
        if ( columnsDefinition.containsKey( name ) )
        {
            Map<Integer, ColumnDefinition> colDefinition = new HashMap<>();
            if ( !isColumnDefinitionEmpty( name ) )
            {
                colDefinition = columnsDefinition.get( name );
            }

            ColumnDefinition definition = colDefinition.get( column );
            if ( definition != null )
            {
                colDefinition.put( column,
                                   new ColumnDefinition( definition.getHeader(), definition.getFormatting(),
                                                         definition.getExportingColumnType(), precision, scale ) );
                columnsDefinition.put( name, colDefinition );
            }
        }
    }

    public void setColumnDefinition( String name, Map<Integer, ColumnDefinition> definition )
    {
        columnsDefinition.put( name, definition );
    }

    public void setColumnDefinition( String name, int column, String columnHeader )
    {
        if ( columnsDefinition.containsKey( name ) )
        {
            Map<Integer, ColumnDefinition> definition = new HashMap<>();
            if ( !isColumnDefinitionEmpty( name ) )
            {
                definition = columnsDefinition.get( name );
            }
            definition.put( column, new ColumnDefinition( columnHeader, null, null ) );
            columnsDefinition.put( name, definition );
        }
    }

    public void setColumnDefinition( String name, int column, String columnHeader, Formatting formatting )
    {
        if ( formatting != null )
        {
            formatting.setColumn( column );
        }

        if ( columnsDefinition.containsKey( name ) )
        {
            Map<Integer, ColumnDefinition> definition = new HashMap<>();
            if ( !isColumnDefinitionEmpty( name ) )
            {
                definition = columnsDefinition.get( name );
            }
            definition.put( column, new ColumnDefinition( columnHeader, formatting, ExportingColumnType.TEXT ) );
            columnsDefinition.put( name, definition );
        }
    }

    public void setColumnDefinition( String name, int column, String columnHeader, Formatting formatting,
                                     ExportingColumnType exportedColumnType )
    {
        if ( formatting != null )
        {
            formatting.setColumn( column );
        }

        if ( columnsDefinition.containsKey( name ) )
        {
            Map<Integer, ColumnDefinition> definition = new HashMap<>();
            if ( !isColumnDefinitionEmpty( name ) )
            {
                definition = columnsDefinition.get( name );
            }
            definition.put( column, new ColumnDefinition( columnHeader, formatting, exportedColumnType ) );
            columnsDefinition.put( name, definition );
        }
    }

    public void setColumnDefinition( String name, int column, String columnHeader, Formatting formatting,
                                     ExportingColumnType exportedColumnType, int length )
    {
        if ( formatting != null )
        {
            formatting.setColumn( column );
        }

        if ( columnsDefinition.containsKey( name ) )
        {
            Map<Integer, ColumnDefinition> definition = new HashMap<>();
            if ( !isColumnDefinitionEmpty( name ) )
            {
                definition = columnsDefinition.get( name );
            }
            definition.put( column, new ColumnDefinition( columnHeader, formatting, exportedColumnType, length ) );
            columnsDefinition.put( name, definition );
        }
    }

    public void setColumnDefinition( String name, int column, String columnHeader, Formatting formatting,
                                     ExportingColumnType exportedColumnType, int precision, int scale )
    {
        if ( formatting != null )
        {
            formatting.setColumn( column );
        }

        if ( columnsDefinition.containsKey( name ) )
        {
            Map<Integer, ColumnDefinition> definition = new HashMap<>();
            if ( !isColumnDefinitionEmpty( name ) )
            {
                definition = columnsDefinition.get( name );
            }
            definition.put( column,
                            new ColumnDefinition( columnHeader, formatting, exportedColumnType, scale, precision ) );
            columnsDefinition.put( name, definition );
        }
    }

    public Object getDefaultValueForHiddenAndRemovedColumns()
    {
        return defaultValueForHiddenAndRemovedColumns;
    }

    public void setDefaultValueForHiddenAndRemovedColumns( Object defaultValueForHiddenAndRemovedColumns )
    {
        if ( defaultValueForHiddenAndRemovedColumns != null && !defaultValueForHiddenAndRemovedColumns.toString()
                                                                                                      .isEmpty()
            && !" ".equals( defaultValueForHiddenAndRemovedColumns.toString()
                                                                  .trim() ) )
        {
            this.defaultValueForHiddenAndRemovedColumns = defaultValueForHiddenAndRemovedColumns;
        }

    }

    public int getWindowRefreshDelay()
    {
        return windowRefreshDelay;
    }

    public void setWindowRefreshDelay( int windowRefreshDelay )
    {
        this.windowRefreshDelay = windowRefreshDelay;
    }

    public boolean isExcludeBaseDataEnabled()
    {
        return excludeBaseDataEnabled;
    }

    public void setExcludeBaseDataEnabled( boolean excludeBaseDataEnabled )
    {
        this.excludeBaseDataEnabled = excludeBaseDataEnabled;
    }

    public boolean isSelectHiddenColumnsEnabled()
    {
        return selectHiddenColumnsEnabled;
    }

    public void setSelectHiddenColumnsEnabled( boolean selectHiddenColumnsEnabled )
    {
        this.selectHiddenColumnsEnabled = selectHiddenColumnsEnabled;
    }

    public boolean isRemovedOrdersFromColumnsEnabled()
    {
        return removeOrdersFromColumnsEnabled;
    }

    public void setRemoveOrdersFromColumnsEnabled( boolean removeOrdersFromColumnsEnabled )
    {
        this.removeOrdersFromColumnsEnabled = removeOrdersFromColumnsEnabled;
    }

    public boolean isPrepareDataWithOrdersEnabled()
    {
        return prepareDataWithOrdersEnabled;
    }

    public void setPrepareDataWithOrdersEnabled( boolean prepareDataWithOrdersEnabled )
    {
        this.prepareDataWithOrdersEnabled = prepareDataWithOrdersEnabled;
    }

    public boolean isPrepareDataWithHiddenColumnsEnabled()
    {
        return prepareDataWithHiddenColumnsEnabled;
    }

    public void setPrepareDataWithHiddenColumnsEnabled( boolean prepareDataWithHiddenColumnsEnabled )
    {
        this.prepareDataWithHiddenColumnsEnabled = prepareDataWithHiddenColumnsEnabled;
    }

    public boolean isPrepareDataWithRemovedColumnsEnabled()
    {
        return prepareDataWithRemovedColumnsEnabled;
    }

    public void setPrepareDataWithRemovedColumnsEnabled( boolean prepareDataWithRemovedColumnsEnabled )
    {
        this.prepareDataWithRemovedColumnsEnabled = prepareDataWithRemovedColumnsEnabled;
    }

    public boolean isPrepareDataWithOriginalValuesForHiddenAndRemovedColumnsEnabled()
    {
        return prepareDataWithOriginalValuesForHiddenAndRemovedColumnsEnabled;
    }

    public void setPrepareDataWithOriginalValuesForHiddenAndRemovedColumnsEnabled( boolean prepareDataWithOriginalValuesForHiddenAndRemovedColumnsEnabled )
    {
        this.prepareDataWithOriginalValuesForHiddenAndRemovedColumnsEnabled =
            prepareDataWithOriginalValuesForHiddenAndRemovedColumnsEnabled;
    }

    public boolean isExportCsvFileEnabled()
    {
        return exportCsvFileEnabled;
    }

    public void setExportCsvFileEnabled( boolean exportCsvFileEnabled )
    {
        this.exportCsvFileEnabled = exportCsvFileEnabled;
    }

    public boolean isExportMdbFileEnabled()
    {
        return exportMdbFileEnabled;
    }

    public void setExportMdbFileEnabled( boolean exportMdbFileEnabled )
    {
        this.exportMdbFileEnabled = exportMdbFileEnabled;
    }

    public boolean isExportPdfFileEnabled()
    {
        return exportPdfFileEnabled;
    }

    public void setExportPdfFileEnabled( boolean exportPdfFileEnabled )
    {
        // TODO: enable
        // this.exportPdfFileEnabled = exportPdfFileEnabled;
    }

    public boolean isExportXlsFileEnabled()
    {
        return exportXlsFileEnabled;
    }

    public void setExportXlsFileEnabled( boolean exportXlsFileEnabled )
    {
        this.exportXlsFileEnabled = exportXlsFileEnabled;
    }

    public boolean isFormattedDataEnabled()
    {
        return formattedDataEnabled;
    }

    public void setFormattedDataEnabled( boolean formattedDataEnabled )
    {
        this.formattedDataEnabled = formattedDataEnabled;
    }

    public boolean isRowIndexEnabled()
    {
        return rowIndexEnabled;
    }

    public void setRowIndexEnabled( boolean rowIndexEnabled )
    {
        this.rowIndexEnabled = rowIndexEnabled;
    }

    public boolean isTimestampEnabled()
    {
        return timestampEnabled;
    }

    public void setTimestampEnabled( boolean timestampEnabled )
    {
        this.timestampEnabled = timestampEnabled;
    }

    public String getTimestampFormat()
    {
        return timestampFormat;
    }

    public void setTimestampFormat( String timestampFormat )
    {
        this.timestampFormat = timestampFormat;
    }

    public String getFilename()
    {
        return fileName;
    }

    public void setFilename( String filename )
    {
        this.fileName = filename;
    }

    public String getTableName()
    {
        return tableName;
    }

    public void setTableName( String tableName )
    {
        this.tableName = tableName;
    }

    public String getCsvColumnDelimiter()
    {
        return csvColumnDelimiter;
    }

    public void setCsvColumnDelimiter( String csvColumnDelimiter )
    {
        this.csvColumnDelimiter = csvColumnDelimiter;
    }

    public MsAccessVersion getMsAccessVersion()
    {
        return msAccessVersion;
    }

    public void setMsAccessVersion( MsAccessVersion msAccessVersion )
    {
        this.msAccessVersion = msAccessVersion;
    }

    public MsExcelVersion getMsExcelVersion()
    {
        return msExcelVersion;
    }

    public void setMsExcelVersion( MsExcelVersion msExcelVersion )
    {
        this.msExcelVersion = msExcelVersion;
    }

    public Charset getCharset()
    {
        return charset;
    }

    public void setCharset( Charset charset )
    {
        this.charset = charset;
    }

    public CompressionType getCompressionType()
    {
        return compressionType;
    }

    public void setCompression( CompressionType compressionType )
    {
        this.compressionType = compressionType;
    }

    public Map<String, String> getData()
    {
        Map<String, String> data = new HashMap<>();
        data.put( "exportCsvFileEnabled", Boolean.toString( isExportCsvFileEnabled() ) );
        data.put( "exportMdbFileEnabled", Boolean.toString( isExportMdbFileEnabled() ) );
        data.put( "exportPdfFileEnabled", Boolean.toString( isExportPdfFileEnabled() ) );
        data.put( "exportXlsFileEnabled", Boolean.toString( isExportXlsFileEnabled() ) );
        return data;
    }

}
