/**
 * Copyright (C) 2012-2016  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dtable.windows.exporter;

import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.ss.usermodel.CellStyle;
import org.apache.poi.ss.usermodel.CreationHelper;
import org.apache.poi.ss.usermodel.RichTextString;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.ss.util.WorkbookUtil;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.dussan.vaadin.dtable.DTable;
import org.dussan.vaadin.dtable.config.formatting.Formatting;
import org.dussan.vaadin.dtable.translations.TranslatedWords;
import org.dussan.vaadin.dtable.utils.ErrorUtil;
import org.dussan.vaadin.dtable.windows.exporter.base.Exporter;
import org.dussan.vaadin.dtable.windows.exporter.type.ExportedType;
import org.dussan.vaadin.dtable.windows.exporter.type.ExportingColumnType;
import org.dussan.vaadin.dtable.windows.exporter.version.MsExcelVersion;

public class XlsExporter
    extends Exporter
{

    private static final long serialVersionUID = -6601924322990231070L;

    private int columnStep = MsExcelVersion.V2007.getMaxColumns();

    private List<String> sheetNames = null;

    private Map<Integer, Short> cellStyleIndex = null;

    private FileOutputStream fileOutputStream = null;

    private Workbook workbook = null;

    public XlsExporter( DTable dTable )
    {
        super( dTable );

        cellStyleIndex = new HashMap<>();
        sheetNames = new ArrayList<>();
        columnStep = getDTable().getTableExporterConfig()
                                .getMsExcelVersion()
                                .getMaxColumns();

        setCaption( getDTable().getTranslation()
                               .get( getDTable().getTableConfig()
                                                .getLocale(),
                                     TranslatedWords.EXPORTING_DATA_TO_MS_EXCEL_FILE ) );
        setExportedType( ExportedType.MS_EXCEL );

        if ( getDTable().getTableExporterConfig()
                        .getMsExcelVersion()
                        .equals( MsExcelVersion.V2007 ) )
        {
            workbook = new XSSFWorkbook();
        }
        else
        {
            workbook = new HSSFWorkbook();
        }
    }

    private void createSheetsHeader( String name )
    {
        try
        {
            int columnsDefinitionSize = getDTable().getTableExporterConfig()
                                                   .getColumnDefinition( name )
                                                   .size();
            String partialSheetName = getDTable().getTranslation()
                                                 .get( getDTable().getTableConfig()
                                                                  .getLocale(),
                                                       TranslatedWords.MS_EXCEL_PARTIAL_SHEET_NAME );

            List<Integer> columns = new ArrayList<>( getDTable().getTableExporterConfig()
                                                                .getColumnDefinition( name )
                                                                .keySet() );
            Collections.sort( columns );

            int colStep = columnStep;
            int columnStepIndexAdapter = 0;
            if ( getDTable().getTableExporterConfig()
                            .isRowIndexEnabled() )
            {
                colStep -= 1;
                columnStepIndexAdapter = 1;
            }

            sheetNames.clear();
            CreationHelper creationHelper = workbook.getCreationHelper();
            for ( int sheetsCounter = 0; (float) sheetsCounter < (float) columnsDefinitionSize
                / colStep; sheetsCounter++ )
            {
                String sheetName = getDTable().getTableExporterConfig()
                                              .getBaseName()
                                              .equals( name ) ? getDTable().getTableExporterConfig()
                                                                           .getTableName()
                                                              : name;
                sheetName += columnsDefinitionSize <= colStep ? "" : ( partialSheetName + ( sheetsCounter + 1 ) );
                sheetName = WorkbookUtil.createSafeSheetName( sheetName );
                Sheet sheet = workbook.createSheet( sheetName );

                Row row = sheet.createRow( 0 );
                if ( getDTable().getTableExporterConfig()
                                .isRowIndexEnabled() )
                {
                    row.createCell( 0 )
                       .setCellValue( "#" );
                }

                int startColumn = colStep * sheetsCounter;
                for ( int column = colStep * sheetsCounter; column < colStep * ( sheetsCounter + 1 )
                    && column < columnsDefinitionSize; column++ )
                {
                    int headerColumn = columns.get( column );
                    String header = getDTable().getTableExporterConfig()
                                               .getColumnDefinitionHeader( name, headerColumn );
                    row.createCell( column - startColumn + columnStepIndexAdapter )
                       .setCellValue( header );

                    // column style
                    CellStyle cellStyle = workbook.createCellStyle();
                    Formatting formatting = getDTable().getTableExporterConfig()
                                                       .getColumnDefinitionFormatting( name, headerColumn );
                    ExportingColumnType columnType = getDTable().getTableExporterConfig()
                                                                .getColumnDefinitionExportingType( name, headerColumn );
                    if ( ( getDTable().getTableExporterConfig()
                                      .isFormattedDataEnabled()
                        || ExportingColumnType.DATE.equals( columnType )
                        || ExportingColumnType.TIMESTAMP.equals( columnType ) ) && formatting != null
                        && formatting.getPattern() != null )
                    {
                        cellStyle.setDataFormat( creationHelper.createDataFormat()
                                                               .getFormat( formatting.getPattern() ) );
                    }
                    cellStyleIndex.put( headerColumn, cellStyle.getIndex() );
                }
                sheetNames.add( sheetName );
            }
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            setExportedStatus( EXPORT_ERROR, e.getLocalizedMessage() );
        }
    }

    @Override
    protected void exportDataIntoFile()
    {
        try
        {
            if ( 0 < getDTable().getTableExporterConfig()
                                .getColumnsDefinitionName().length )
            {
                float completeStep = 0f;
                float step = 1f / getDTable().getTableExporterConfig()
                                             .getColumnsDefinitionName().length;

                setExportedStatus( EXPORT_BEGIN );
                if ( openFile( getDTable().getTableExporterConfig()
                                          .getFilename() ) )
                {
                    for ( String name : getDTable().getTableExporterConfig()
                                                   .getColumnsDefinitionName() )
                    {
                        if ( prepareExportedData( name ) && canContinue() )
                        {
                            writeToFile( name, completeStep, step );
                            completeStep += step;
                        }
                        else
                        {
                            setExportedStatus( EXPORT_ERROR );
                        }

                        // check status
                        if ( !canContinue() )
                        {
                            closeFile();
                            if ( getExportedStatus() == EXPORT_STOP )
                            {
                                setExportedStatus( EXPORT_STOPPED );
                            }
                            return;
                        }
                    }
                }
                else
                {
                    closeFile();
                    setExportedStatus( EXPORT_ERROR );
                    return;
                }

                closeFile();
                setExportedStatus( EXPORT_COMPLETE );
            }
            else
            {
                setExportedStatus( EXPORT_ERROR, getDTable().getTranslation()
                                                            .get( getDTable().getTableConfig()
                                                                             .getLocale(),
                                                                  TranslatedWords.NO_DATA_FOR_EXPORT ) );
            }
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            setExportedStatus( EXPORT_ERROR, e.getLocalizedMessage() );
        }
    }

    @Override
    protected boolean openFile( String name )
    {
        try
        {
            fileOutputStream = new FileOutputStream( getTempFile( name ), true );
            return true;
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            setExportedStatus( EXPORT_ERROR, e.getLocalizedMessage() );
            return false;
        }
    }

    @Override
    protected void writeToFile( String name, float completeStep, float step )
    {
        boolean errorOnMaxAlowedRows = false;

        try
        {
            int colStep = columnStep;
            int columnStepIndexAdapter = 0;
            if ( getDTable().getTableExporterConfig()
                            .isRowIndexEnabled() )
            {
                colStep -= 1;
                columnStepIndexAdapter = 1;
            }

            createSheetsHeader( name );
            if ( getExportedStatus() != EXPORT_ERROR )
            {
                long allRows = getAllRows( name );
                int allColumns = getAllColumns( name );
                if ( 0 < allColumns && 0 < allRows && 0 < getExportedStep( name ) )
                {
                    int exportedStep = allRows < getExportedStep( name ) ? (int) allRows : getExportedStep( name );
                    List<Integer> columns = new ArrayList<>( getDTable().getTableExporterConfig()
                                                                        .getColumnDefinition( name )
                                                                        .keySet() );
                    Collections.sort( columns );

                    for ( int mainRow = 0; mainRow < allRows; mainRow += exportedStep )
                    {
                        int rows =
                            (int) ( ( mainRow + exportedStep ) <= allRows ? exportedStep : ( allRows - mainRow ) );
                        List<Object> rowsData = Arrays.asList( getOriginalData( name, mainRow, rows ) );

                        if ( getDTable().getTableExporterConfig()
                                        .getMsExcelVersion()
                                        .getMaxRows() <= ( mainRow + exportedStep ) )
                        {
                            errorOnMaxAlowedRows = true;
                            rows = getDTable().getTableExporterConfig()
                                              .getMsExcelVersion()
                                              .getMaxRows()
                                - mainRow - 1;

                            // reset rows if main row is greater than max allowed columns
                            if ( getDTable().getTableExporterConfig()
                                            .getMsExcelVersion()
                                            .getMaxRows() < mainRow )
                            {
                                rows = 0;
                            }
                        }

                        for ( int sheetName = 0; sheetName < sheetNames.size(); sheetName++ )
                        {
                            int fromColumn = colStep * sheetName;
                            int toColumn = colStep * ( sheetName + 1 );
                            toColumn = Math.min( toColumn, allColumns );
                            Sheet sheet = workbook.getSheet( sheetNames.get( sheetName ) );

                            for ( int row = 0; row < rows; row++ )
                            {
                                List<Object> data =
                                    rowsData.subList( allColumns * row + fromColumn, allColumns * row + toColumn );
                                Row sheetRow = sheet.createRow( 1 + mainRow + row );
                                if ( getDTable().getTableExporterConfig()
                                                .isRowIndexEnabled() )
                                {
                                    sheetRow.createCell( 0 )
                                            .setCellValue( mainRow + row + 1 );
                                }

                                for ( int column = 0; column < data.size(); column++ )
                                {
                                    Object cellValue = data.get( column );
                                    if ( cellValue != null && !cellValue.toString()
                                                                        .isEmpty() )
                                    {
                                        int headerColumn = columns.get( fromColumn + column );
                                        ExportingColumnType columnType = getDTable().getTableExporterConfig()
                                                                                    .getColumnDefinitionExportingType( name,
                                                                                                                       headerColumn );
                                        switch ( columnType )
                                        {
                                            case BOOLEAN:
                                                sheetRow.createCell( column + columnStepIndexAdapter )
                                                        .setCellValue( Boolean.parseBoolean( cellValue.toString() ) );
                                                break;
                                            case BIG_DECIMAL:
                                            case BIG_INTEGER:
                                            case DOUBLE:
                                            case FLOAT:
                                            case INTEGER:
                                            case LONG:
                                                sheetRow.createCell( column + columnStepIndexAdapter )
                                                        .setCellValue( Double.parseDouble( cellValue.toString() ) );
                                                break;
                                            case MEMO:
                                                sheetRow.createCell( column + columnStepIndexAdapter )
                                                        .setCellValue( (RichTextString) cellValue );
                                                break;
                                            case DATE:
                                            case TIMESTAMP:
                                                sheetRow.createCell( column + columnStepIndexAdapter )
                                                        .setCellValue( (Date) cellValue );
                                                break;
                                            case TEXT:
                                            default:
                                                sheetRow.createCell( column + columnStepIndexAdapter )
                                                        .setCellValue( cellValue.toString() );
                                                break;
                                        }
                                        sheetRow.getCell( column + columnStepIndexAdapter )
                                                .setCellStyle( workbook.getCellStyleAt( cellStyleIndex.get( headerColumn ) ) );
                                    }
                                }

                                // exports was stopped
                                if ( !canContinue() )
                                {
                                    setExportedStatus( EXPORT_STOPPED );
                                    return;
                                }
                            }

                            // only for MS Excel 2007+
                            if ( getDTable().getTableExporterConfig()
                                            .getMsExcelVersion()
                                            .equals( MsExcelVersion.V2007 ) )
                            {
                                openFile( name );
                                workbook.write( fileOutputStream );
                                fileOutputStream.close();
                            }
                        }
                        setExportedStatus( ( step * mainRow / allRows ) + completeStep );
                    }
                }
                else
                {
                    setExportedStatus( EXPORT_ERROR, getDTable().getTranslation()
                                                                .get( getDTable().getTableConfig()
                                                                                 .getLocale(),
                                                                      TranslatedWords.NO_DATA_FOR_EXPORT ) );
                }
            }

            // only for MS Excel 2003-
            if ( getDTable().getTableExporterConfig()
                            .getMsExcelVersion()
                            .equals( MsExcelVersion.V2003 ) )
            {
                workbook.write( fileOutputStream );
                fileOutputStream.flush();
            }
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            setExportedStatus( EXPORT_ERROR, e.getLocalizedMessage() );
        }

        if ( errorOnMaxAlowedRows )
        {
            new ErrorUtil( getDTable().getTranslation()
                                      .get( getDTable().getTableConfig()
                                                       .getLocale(),
                                            TranslatedWords.INFO ),
                           getDTable().getTranslation()
                                      .get( getDTable().getTableConfig()
                                                       .getLocale(),
                                            TranslatedWords.SOME_DATA_ARE_NOT_EXPORTED_BECAUSE_THEY_EXCEEDED_THE_ALLOWED_QUOTA_OF_ROWS_WHICH_IS_ALLOWED_IN_MS_EXCEL ) ).show();
        }
    }

    @Override
    protected void closeFile()
    {
        try
        {
            workbook.close();
            fileOutputStream.close();
        }
        catch ( Exception e )
        {
            Logger.getLogger( DTable.class.getName() )
                  .log( Level.INFO, e.getLocalizedMessage(), e );
            setExportedStatus( EXPORT_ERROR, e.getLocalizedMessage() );
        }
    }

}
