/**
 * Copyright (C) 2012-2016  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dtable.windows.hide;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import org.dussan.vaadin.dtable.DTable;
import org.dussan.vaadin.dtable.translations.TranslatedWords;
import org.dussan.vaadin.dtable.utils.ErrorUtil;

import com.vaadin.data.Property.ValueChangeEvent;
import com.vaadin.data.Property.ValueChangeListener;
import com.vaadin.event.FieldEvents.TextChangeEvent;
import com.vaadin.event.FieldEvents.TextChangeListener;
import com.vaadin.server.FontAwesome;
import com.vaadin.ui.Alignment;
import com.vaadin.ui.Button;
import com.vaadin.ui.Button.ClickEvent;
import com.vaadin.ui.Button.ClickListener;
import com.vaadin.ui.HorizontalLayout;
import com.vaadin.ui.JavaScript;
import com.vaadin.ui.Label;
import com.vaadin.ui.ListSelect;
import com.vaadin.ui.TextField;
import com.vaadin.ui.UI;
import com.vaadin.ui.VerticalLayout;
import com.vaadin.ui.Window;
import com.vaadin.ui.themes.ValoTheme;

public class HideUnhideColumns
    extends Window
{

    private static final long serialVersionUID = 3771051434575477411L;

    private int actionDelay = 50;

    private int initDelay = 200;

    private String frozenColumnDescription = null;

    private List<String> disabledHiddenColumns = null;

    private List<String> disabledUnhiddenColumns = null;

    private List<String> hiddenUnfilteredColumns = null;

    private Button closeButton = null;

    private Button hideButton = null;

    private Button okButton = null;

    private Button unhideButton = null;

    private ListSelect hiddenColumns = null;

    private ListSelect unhiddenColumns = null;

    private TextField hiddenFilter = null;

    private TextField unhiddenFilter = null;

    private DTable dTable = null;

    /**
     * Creates a new instance.
     */
    private HideUnhideColumns()
    {
        disabledHiddenColumns = new ArrayList<>();
        disabledUnhiddenColumns = new ArrayList<>();
    }

    /**
     * Creates a new instance with dTable object.
     * 
     * @param dTable dDable instance
     */
    public HideUnhideColumns( DTable dTable )
    {
        this();
        this.dTable = dTable;

        frozenColumnDescription = " (" + getDTable().getTranslation()
                                                    .get( getDTable().getTableConfig()
                                                                     .getLocale(), TranslatedWords.FROZEN ) + ")";

        init();
    }

    /**
     * Get dTable instance.
     * 
     * @return dTable instance
     */
    private DTable getDTable()
    {
        return dTable;
    }

    /**
     * Closes window.
     */
    private void closeWindow()
    {
        setVisible( false );
        UI.getCurrent()
          .removeWindow( this );
    }

    /**
     * Javascript code which disable specified columns in list.
     * 
     * @param id lists id
     * @param disabledColumns disabled columns
     * @param delay action delay in ms
     */
    private void setDisabledColumns( String id, String disabledColumns, int delay )
    {
        if ( id != null && !id.isEmpty() && disabledColumns != null && !disabledColumns.isEmpty() && 0 < delay )
        {
            StringBuilder jsCode = new StringBuilder();
            jsCode.append( "setTimeout( function(){" );
            jsCode.append( "    var disabledColumns = " + disabledColumns + ";" );
            jsCode.append( "    var options = document.getElementById('" + id + "').children[0].options;" );
            jsCode.append( "    for( var column = 0; column < disabledColumns.length; column++ ){" );
            jsCode.append( "        for( var option = 0; option < options.length; option++ ){" );
            jsCode.append( "            if( disabledColumns[column] === options[option].text ){" );
            jsCode.append( "                options[option].disabled = true;" );
            jsCode.append( "            }" );
            jsCode.append( "        }" );
            jsCode.append( "    }" );
            jsCode.append( "}, " + delay + " );" );
            JavaScript.getCurrent()
                      .execute( jsCode.toString() );
        }
    }

    /**
     * Get column header index from header value.
     * 
     * @param header header value
     * @return column header index
     */
    private int getColumnHeaderIndex( Object header )
    {
        String headerString = header.toString();
        if ( headerString.endsWith( frozenColumnDescription ) )
        {
            headerString = headerString.substring( 0, headerString.lastIndexOf( frozenColumnDescription ) );
        }
        return getDTable().getTableProperties()
                          .getColumnHeaderIndex( headerString );
    }

    /**
     * Get new column size for frozen table.
     * 
     * @return new column size for frozen table
     */
    private int getNewFrozenColumnsSize()
    {
        int columns = 0;
        for ( Object header : unhiddenColumns.getItemIds() )
        {
            int column = getColumnHeaderIndex( header );
            if ( getDTable().getTableProperties()
                            .isFrozenColumn( column ) && !getDTable().getTableProperties()
                                                                     .isHiddenColumnFrozen( column )
                && !getDTable().getTableProperties()
                               .isRemovedColumn( column ) )
            {
                columns += 1;
            }
        }
        return columns;
    }

    /**
     * Get new column size for table.
     * 
     * @return new column size
     */
    private int getNewUnfrozenColumnsSize()
    {
        int columns = 0;
        for ( Object header : unhiddenColumns.getItemIds() )
        {
            int column = getColumnHeaderIndex( header );
            if ( !getDTable().getTableProperties()
                             .isFrozenColumn( column ) && !getDTable().getTableProperties()
                                                                      .isHiddenColumn( column )
                && !getDTable().getTableProperties()
                               .isRemovedColumn( column ) )
            {
                columns += 1;
            }
        }
        return columns;
    }

    /**
     * Add disabled hidden column to list which is later used in javascript code.
     * 
     * @param header column header
     */
    private void addDisabledHiddenColumn( String header )
    {

        if ( header != null && !header.trim()
                                      .isEmpty() )
        {
            disabledHiddenColumns.add( header );
        }
    }

    /**
     * Remove disabled hidden column from list which is later used in javascript code.
     * 
     * @param header column header
     */
    private void removeDisabledHiddenColumn( String header )
    {
        if ( header != null && !header.trim()
                                      .isEmpty() )
        {
            disabledHiddenColumns.remove( header );
        }
    }

    /**
     * Get string from disabled hidden columns array.
     * 
     * @return string from disabled hidden columns array
     */
    private String getDisabledHiddenColumnsArrayString()
    {
        StringBuilder string = new StringBuilder();
        if ( !disabledHiddenColumns.isEmpty() )
        {
            for ( String header : disabledHiddenColumns )
            {
                string.append( ",'" + header + "'" );
            }

            if ( 0 < string.length() )
            {
                string = new StringBuilder( "[" + string.substring( 1 ) + "]" );
            }
        }
        return string.toString();
    }

    /**
     * Add disabled unhidden column to list which is later used in javascript code.
     * 
     * @param header column header
     */
    private void addDisabledUnhiddenColumn( String header )
    {
        if ( header != null && !header.trim()
                                      .isEmpty() )
        {
            disabledUnhiddenColumns.add( header );
        }
    }

    /**
     * Remove disabled unhidden column from list which is later used in javascript code.
     * 
     * @param header column header
     */
    private void removeDisabledUnhiddenColumn( String header )
    {
        if ( header != null && !header.trim()
                                      .isEmpty() )
        {
            disabledUnhiddenColumns.remove( header );
        }
    }

    /**
     * Get string from disabled unhidden columns array.
     * 
     * @return string from disabled unhidden columns array
     */
    private String getDisabledUnhiddenColumnsArrayString()
    {
        StringBuilder string = new StringBuilder();
        if ( !disabledUnhiddenColumns.isEmpty() )
        {
            for ( String header : disabledUnhiddenColumns )
            {
                string.append( ",'" + header + "'" );
            }

            if ( 0 < string.length() )
            {
                string = new StringBuilder( "[" + string.substring( 1 ) + "]" );
            }
        }
        return string.toString();
    }

    /**
     * Check if button for unhidden columns may be enabled.
     */
    private void checkUnhiddenButton()
    {
        boolean buttonEnabled = false;
        for ( Object header : hiddenColumns.getItemIds() )
        {
            int column = getColumnHeaderIndex( header );
            if ( getDTable().getTableConfig()
                            .canColumnUnhide( column ) && !getDTable().getTableProperties()
                                                                      .isRemovedColumn( column ) )
            {
                buttonEnabled = true;
                break;
            }
        }
        unhideButton.setEnabled( buttonEnabled );
    }

    /**
     * Check if button for hidden columns may be enabled.
     */
    private void checkHideButton()
    {
        boolean buttonEnabled = false;
        for ( Object header : unhiddenColumns.getItemIds() )
        {
            int column = getColumnHeaderIndex( header );
            if ( getDTable().getTableConfig()
                            .canColumnHide( column ) && !getDTable().getTableProperties()
                                                                    .isRemovedColumn( column ) )
            {
                buttonEnabled = true;
                break;
            }
        }
        hideButton.setEnabled( buttonEnabled );
    }

    /**
     * Fill lists with hidden and unhidden table columns.
     */
    private void fillHiddenAndUnhiddenLists()
    {
        hiddenUnfilteredColumns = new ArrayList<>();
        unhiddenColumns.removeAllItems();
        hiddenColumns.removeAllItems();
        for ( int column = 0; column < getDTable().getTableProperties()
                                                  .getAllColumns(); column++ )
        {
            String header = getDTable().getTableProperties()
                                       .getColumnHeader( column );

            if ( getDTable().getTableProperties()
                            .isFrozenColumn( column ) || getDTable().getTableProperties()
                                                                    .isHiddenColumnFrozen( column ) )
            {
                header += frozenColumnDescription;
            }

            if ( !getDTable().getTableProperties()
                             .isHiddenColumn( column ) && !getDTable().getTableProperties()
                                                                      .isRemovedColumn( column ) )
            {
                unhiddenColumns.addItem( header );

                if ( !getDTable().getTableConfig()
                                 .canColumnHide( column ) )
                {
                    addDisabledUnhiddenColumn( header );
                }
            }

            if ( getDTable().getTableProperties()
                            .isHiddenColumn( column ) && !getDTable().getTableProperties()
                                                                     .isRemovedColumn( column ) )
            {
                hiddenColumns.addItem( header );
                hiddenUnfilteredColumns.add( header );

                if ( !getDTable().getTableConfig()
                                 .canColumnUnhide( column ) )
                {
                    addDisabledHiddenColumn( header );
                }
            }
        }

        // disable unhide button if hidden table cannot be empty
        checkUnhiddenButton();

        // disable hide button if unhidden list is empty
        checkHideButton();
    }

    /**
     * Initialization procedure.
     */
    private void init()
    {
        int rows = 10;

        center();
        setClosable( false );
        setDraggable( false );
        setModal( true );
        setResizable( false );
        setWidth( 70, Unit.PERCENTAGE );
        setCaption( getDTable().getTranslation()
                               .get( getDTable().getTableConfig()
                                                .getLocale(), TranslatedWords.HIDE_UNHIDE_COLUMNS ) );

        // unhidden filter
        unhiddenFilter = new TextField( getDTable().getTranslation()
                                                   .get( getDTable().getTableConfig()
                                                                    .getLocale(), TranslatedWords.AVAILABLE_COLUMNS ) );
        unhiddenFilter.addStyleName( ValoTheme.TEXTFIELD_SMALL );
        unhiddenFilter.addTextChangeListener( new TextChangeListener()
        {
            private static final long serialVersionUID = -8451981510521085283L;

            @Override
            public void textChange( TextChangeEvent event )
            {
                String value = event.getText();

                unhiddenColumns.removeAllItems();
                for ( int column = 0; column < getDTable().getTableProperties()
                                                          .getAllColumns(); column++ )
                {
                    String header = getDTable().getTableProperties()
                                               .getColumnHeader( column );

                    if ( getDTable().getTableProperties()
                                    .isHiddenColumn( column ) )
                    {
                        header += frozenColumnDescription;
                    }

                    if ( ( !getDTable().getTableProperties()
                                       .isFrozenColumn( column ) || getDTable().getTableProperties()
                                                                               .isHiddenColumn( column ) )
                        && !getDTable().getTableProperties()
                                       .isHiddenColumnFrozen( column ) && !getDTable().getTableProperties()
                                                                                      .isRemovedColumn( column )
                        && !hiddenColumns.containsId( header ) && ( value.isEmpty() || header.contains( value ) ) )
                    {
                        unhiddenColumns.addItem( header );
                    }
                }

                // disable unhidden columns which can not hide
                setDisabledColumns( unhiddenColumns.getId(), getDisabledUnhiddenColumnsArrayString(), actionDelay );

                // disable hide button if unhidden list is empty
                checkHideButton();
            }
        } );

        // hidden filter
        hiddenFilter = new TextField( getDTable().getTranslation()
                                                 .get( getDTable().getTableConfig()
                                                                  .getLocale(), TranslatedWords.HIDDEN_COLUMNS ) );
        hiddenFilter.addStyleName( ValoTheme.TEXTFIELD_SMALL );
        hiddenFilter.addTextChangeListener( new TextChangeListener()
        {
            private static final long serialVersionUID = 8593972777238279530L;

            @Override
            public void textChange( TextChangeEvent event )
            {
                String value = event.getText();

                hiddenColumns.removeAllItems();
                for ( String header : hiddenUnfilteredColumns )
                {
                    if ( value.isEmpty() || ( !value.isEmpty() && header.contains( value ) ) )
                    {
                        hiddenColumns.addItem( header );
                    }
                }

                // disable hidden columns which can not unhide
                setDisabledColumns( hiddenColumns.getId(), getDisabledHiddenColumnsArrayString(), actionDelay );

                // disable unhide button if hidden table cannot be empty
                checkUnhiddenButton();
            }
        } );

        // unhidden columns
        unhiddenColumns = new ListSelect();
        unhiddenColumns.setId( "unhiddenColumns" );
        unhiddenColumns.setImmediate( true );
        unhiddenColumns.setMultiSelect( true );
        unhiddenColumns.setNullSelectionAllowed( true );
        unhiddenColumns.setRows( rows );
        unhiddenColumns.setWidth( 100, Unit.PERCENTAGE );
        unhiddenColumns.addValueChangeListener( new ValueChangeListener()
        {
            private static final long serialVersionUID = 4691014034226998827L;

            @Override
            public void valueChange( ValueChangeEvent event )
            {
                // disable unhidden columns which can not hide
                setDisabledColumns( unhiddenColumns.getId(), getDisabledUnhiddenColumnsArrayString(), actionDelay );
            }
        } );

        // hidden columns
        hiddenColumns = new ListSelect();
        hiddenColumns.setId( "hiddenColumns" );
        hiddenColumns.setImmediate( true );
        hiddenColumns.setMultiSelect( true );
        hiddenColumns.setNullSelectionAllowed( true );
        hiddenColumns.setRows( rows );
        hiddenColumns.setWidth( 100, Unit.PERCENTAGE );
        hiddenColumns.addValueChangeListener( new ValueChangeListener()
        {
            private static final long serialVersionUID = 4691014034226998827L;

            @Override
            public void valueChange( ValueChangeEvent event )
            {
                // disable hidden columns which can not unhide
                setDisabledColumns( hiddenColumns.getId(), getDisabledHiddenColumnsArrayString(), actionDelay );
            }
        } );

        // unhide button
        unhideButton = new Button( FontAwesome.CHEVRON_LEFT );
        unhideButton.addStyleName( ValoTheme.BUTTON_ICON_ONLY );
        unhideButton.addClickListener( new ClickListener()
        {
            private static final long serialVersionUID = -6384761875661773041L;

            @Override
            public void buttonClick( ClickEvent event )
            {
                Object values = hiddenColumns.getValue();
                if ( values != null )
                {
                    for ( Object value : (Set<?>) values )
                    {
                        int column = getColumnHeaderIndex( value.toString() );

                        hiddenUnfilteredColumns.remove( value );
                        hiddenColumns.removeItem( value );
                        if ( getDTable().getTableConfig()
                                        .canColumnUnfreeze( column ) )
                        {
                            unhiddenColumns.addItem( value );
                        }

                        removeDisabledHiddenColumn( value.toString() );
                        if ( !getDTable().getTableConfig()
                                         .canColumnFreeze( column ) )
                        {
                            addDisabledUnhiddenColumn( value.toString() );
                        }
                    }
                }
                hiddenColumns.select( null );

                // disable unhidden and hidden columns in lists
                setDisabledColumns( unhiddenColumns.getId(), getDisabledUnhiddenColumnsArrayString(), actionDelay );
                setDisabledColumns( hiddenColumns.getId(), getDisabledHiddenColumnsArrayString(), actionDelay );

                // disable unhide button if hidden table cannot be empty
                checkUnhiddenButton();

                // disable hide button if unhidden list is empty
                checkHideButton();
            }
        } );

        // hide button
        hideButton = new Button( FontAwesome.CHEVRON_RIGHT );
        hideButton.addStyleName( ValoTheme.BUTTON_ICON_ONLY );
        hideButton.addClickListener( new ClickListener()
        {
            private static final long serialVersionUID = 4097366270472216209L;

            @Override
            public void buttonClick( ClickEvent event )
            {
                Object values = unhiddenColumns.getValue();
                if ( values != null )
                {
                    for ( Object value : (Set<?>) values )
                    {
                        int column = getDTable().getTableProperties()
                                                .getColumnHeaderIndex( value.toString() );

                        unhiddenColumns.removeItem( value );
                        if ( getDTable().getTableConfig()
                                        .canColumnFreeze( column ) )
                        {
                            hiddenUnfilteredColumns.add( value.toString() );
                            hiddenColumns.addItem( value );

                            removeDisabledUnhiddenColumn( value.toString() );
                            if ( !getDTable().getTableConfig()
                                             .canColumnUnfreeze( column ) )
                            {
                                addDisabledHiddenColumn( value.toString() );
                            }
                        }
                    }
                }
                unhiddenColumns.select( null );

                // disable unhidden and hidden columns in lists
                setDisabledColumns( unhiddenColumns.getId(), getDisabledUnhiddenColumnsArrayString(), actionDelay );
                setDisabledColumns( hiddenColumns.getId(), getDisabledHiddenColumnsArrayString(), actionDelay );

                // disable unhide button if hidden table cannot be empty
                checkUnhiddenButton();

                // disable hide button if unhidden list is empty
                checkHideButton();
            }
        } );

        // "OK" button
        okButton = new Button( getDTable().getTranslation()
                                          .get( getDTable().getTableConfig()
                                                           .getLocale(), TranslatedWords.OK ) );
        okButton.addClickListener( new ClickListener()
        {
            private static final long serialVersionUID = -5941114122010602045L;

            @Override
            public void buttonClick( ClickEvent event )
            {
                List<Integer> columns = new ArrayList<>();
                if ( hiddenColumns.getItemIds() != null )
                {
                    for ( Object header : (Collection<?>) hiddenColumns.getItemIds() )
                    {
                        columns.add( getColumnHeaderIndex( header.toString() ) );
                    }
                }
                Collections.sort( columns );

                if ( getNewUnfrozenColumnsSize() < 1 )
                {
                    new ErrorUtil(
                                   getDTable().getTranslation()
                                              .get( getDTable().getTableConfig()
                                                               .getLocale(), TranslatedWords.ERROR ),
                                   getDTable().getTranslation()
                                              .get( getDTable().getTableConfig()
                                                               .getLocale(),
                                                    TranslatedWords.AT_LEAST_ONE_COLUMN_HAVE_TO_BE_UNFROZEN_AND_UNHIDDEN ) ).show();
                }
                else if ( !getDTable().getTableConfig()
                                      .isEmptyFrozenTableEnabled() && getNewFrozenColumnsSize() < 1 )
                {
                    new ErrorUtil( getDTable().getTranslation()
                                              .get( getDTable().getTableConfig()
                                                               .getLocale(), TranslatedWords.ERROR ),
                                   getDTable().getTranslation()
                                              .get( getDTable().getTableConfig()
                                                               .getLocale(),
                                                    TranslatedWords.AT_LEAST_ONE_COLUMN_HAVE_TO_BE_FROZEN_AND_UNHIDDEN ) ).show();
                }
                else
                {
                    // set/remove hidden columns
                    for ( Integer column = 0; column < getDTable().getTableProperties()
                                                                  .getAllColumns(); column++ )
                    {
                        if ( columns.contains( column ) && getDTable().getTableConfig()
                                                                      .canColumnHide( column )
                            && !getDTable().getTableProperties()
                                           .isHiddenColumn( column ) && !getDTable().getTableProperties()
                                                                                    .isRemovedColumn( column ) )
                        {
                            getDTable().getTableProperties()
                                       .setHiddenColumn( column );
                        }
                        else if ( !columns.contains( column ) && getDTable().getTableConfig()
                                                                            .canColumnUnhide( column )
                            && getDTable().getTableProperties()
                                          .isHiddenColumn( column ) && !getDTable().getTableProperties()
                                                                                   .isRemovedColumn( column ) )
                        {
                            getDTable().getTableProperties()
                                       .removeHiddenColumn( column );
                        }
                    }

                    // close window
                    closeWindow();

                    // set new table width
                    getDTable().getTableProperties()
                               .calculateTableWidth( getDTable().getTableConfig()
                                                                .getColumnsWidth() );

                    // refresh visible data
                    getDTable().refreshVisibleData( "hideUnhideColumns", true );
                }
            }
        } );

        // close button
        closeButton =
            new Button( getDTable().getTranslation()
                                   .get( getDTable().getTableConfig()
                                                    .getLocale(), TranslatedWords.CLOSE ), new ClickListener()
            {
                private static final long serialVersionUID = 5390169876653829863L;

                @Override
                public void buttonClick( ClickEvent event )
                {
                    closeWindow();
                }
            } );

        // explanation label
        Label explanationLabel =
            new Label( getDTable().getTranslation()
                                  .get( getDTable().getTableConfig()
                                                   .getLocale(),
                                        TranslatedWords.COLUMNS_THAT_CAN_NOT_BE_HIDDEN_OR_UNHIDDEN_ARE_DISABLED ) );

        // layouts
        VerticalLayout frozenLayout = new VerticalLayout( hiddenFilter, hiddenColumns );
        frozenLayout.setMargin( false );
        frozenLayout.setSpacing( true );
        hiddenFilter.setWidth( 100, Unit.PERCENTAGE );

        VerticalLayout unfrozenLayout = new VerticalLayout( unhiddenFilter, unhiddenColumns );
        unfrozenLayout.setMargin( false );
        unfrozenLayout.setSpacing( true );
        unhiddenFilter.setWidth( 100, Unit.PERCENTAGE );

        VerticalLayout actionLayout = new VerticalLayout( hideButton, unhideButton );
        actionLayout.setMargin( false );
        actionLayout.setSizeUndefined();
        actionLayout.setSpacing( false );

        HorizontalLayout columnsLayout = new HorizontalLayout( unfrozenLayout, actionLayout, frozenLayout );
        columnsLayout.setComponentAlignment( actionLayout, Alignment.MIDDLE_CENTER );
        columnsLayout.setExpandRatio( unfrozenLayout, 1.0f );
        columnsLayout.setExpandRatio( frozenLayout, 1.0f );
        columnsLayout.setMargin( false );
        columnsLayout.setSpacing( true );
        columnsLayout.setWidth( 100, Unit.PERCENTAGE );

        VerticalLayout mainLayout = new VerticalLayout( columnsLayout, explanationLabel );
        mainLayout.setMargin( true );
        mainLayout.setSpacing( true );
        mainLayout.setWidth( 100, Unit.PERCENTAGE );
        mainLayout.addStyleName( "context" );

        HorizontalLayout buttonsLayout = new HorizontalLayout( okButton, closeButton );
        buttonsLayout.setComponentAlignment( okButton, Alignment.MIDDLE_RIGHT );
        buttonsLayout.setComponentAlignment( closeButton, Alignment.MIDDLE_LEFT );
        buttonsLayout.setMargin( true );
        buttonsLayout.setSpacing( true );

        VerticalLayout windowLayout = new VerticalLayout( mainLayout, buttonsLayout );
        windowLayout.setComponentAlignment( buttonsLayout, Alignment.BOTTOM_CENTER );
        windowLayout.setMargin( false );
        windowLayout.setSpacing( false );
        windowLayout.addStyleName( "v-dtable-dialog" );
        setContent( windowLayout );

        // init fill hidden and unhidden lists
        fillHiddenAndUnhiddenLists();

        // disable unhidden and hidden columns in lists
        setDisabledColumns( unhiddenColumns.getId(), getDisabledUnhiddenColumnsArrayString(), initDelay );
        setDisabledColumns( hiddenColumns.getId(), getDisabledHiddenColumnsArrayString(), initDelay );
    }

}
